"""
Implements the Runqueue class - a class that manages a list of configurations to pass to a run-function.
"""

import logging
import os
import queue
import sys
import tempfile
import threading
import time
import traceback
import typing
from copy import copy, deepcopy
from datetime import datetime
from enum import Enum

import dill
import multiprocess  # NOTE: we use multiprocess instead of multiprocessing because it allows more flexibility in pickling
import multiprocess.managers as managers  # Same here, use multiprocess
import multiprocess.queues
import PySignal

from configurun.configuration.configuration import Configuration

# from PySide6 import QtCore



class RunQueueHasRunningItemsException(Exception):
	"""
	Exception raised when trying to load a queue from file that contains running items when its allow flag is set to
	False.
	"""

class ConfigurationIsFirmException(Exception):
	"""
	Exception raised when trying to change a configuration for an item for which the configuration is firm (for example:
	configurations that are currently running)
	"""

class WorkspaceInUseException(Exception):
	"""
	Exception raised when trying to use a workspace that is already in use by another process or if the last lock was
	not properly removed.
	"""


log = logging.getLogger(__name__)


CONFIGURATION_MODULE_NAME = "configuration_module"
WORKSPACE_RUN_QUEUE_SAVE_NAME = "run_queue_data.rq" #The name of the file in which the run queue should preferably be saved
WORKSPACE_LOCK_FILE_NAME = ".configurun_workspace.lock" #Is put in the workspace folder to indicate it is in use

class LoggerWriter:
	"""
	Simple wrapper that acts as a file-like object and redirects writes to a logger instance
	"""

	def __init__(self, writer):
		#E.g. pass LoggerWriter(log.info) to redirect to log.info
		self._writer = writer
		self._msg = ""

	def write(self, message):
		"""Write the message to the buffer - upon flush, the buffer will be written to the logger"""
		self._msg = self._msg + message
		while '\n' in self._msg:
			pos = self._msg.find('\n')
			self._writer(self._msg[:pos])
			self._msg = self._msg[pos+1:]

	def flush(self):
		"""Flush the buffer"""
		if self._msg != '':
			self._writer(self._msg)
			self._msg = ''

class FileAndQueueHandler(logging.FileHandler):
	"""
	A class that handles logging to a file and to a queue for runQueue items. A thread can then read from the queue
	and emit signals to the main thread to update the console. The queue is used to send log messages to the main thread
	so that they can be displayed in the UI. Simple wrapper around logging.FileHandler and logging.QueueHandler
	"""
	def __init__(self,
	    	item_id : int,
			item_name : str,
			log_queue : typing.Union[queue.Queue,multiprocess.queues.Queue],
	      	log_filename,
			mode='a',
			encoding=None,
			delay=False
		):
		super().__init__(log_filename, mode, encoding, delay)
		self._log_queue = log_queue
		self._last_fs_pos = 0 #Last file system position (in bytes) - used to keep track of where we are in the file
		self._last_fs_pos = self.stream
		self._item_id = item_id
		self._item_name = item_name
		self._filename = log_filename


	def emit(self, record):
		self._last_fs_pos = os.path.getsize(self._filename) #Get the current file size
		super().emit(record)

		#Format the message according to the format of the logger
		msg = self.format(record)
		linesep = os.linesep #Otherwise we will be offset by 1 due to automatic logging to file adding \r\n (2char)
		msg = msg.replace("\n", linesep)#Replace newline char with the os line separator (otherwise we get offset again)
		if self._log_queue is not None:
			self._log_queue.put_nowait(
				(
					self._item_id,
					self._item_name,
					self._filename, #NOTE: we use the file path as the name - this should be unique even across runs
					datetime.now(),
					self._last_fs_pos,
					msg + linesep
				)
			)

#Enum with possible statuses for a run queue item
class RunQueueItemStatus(Enum):
	"""The status of a run queue item, e.g. queued, running, finished etc."""
	#pylint: disable=invalid-name
	Queued = 0 #Waiting to be run (in queue)
	Running = 1 #Currently running
	Finished = 2 #Finished running succesfully
	Stopped = 3 #Stopped by user during running
	Failed = 4 #Failed during execution
	Cancelled = 5 #Removed from the queue



# @dataclass
class RunQueueItem():
	"""
	A class representing a single item in the run queue
	"""
	def __init__(self,
		  	item_id : int,
		  	name : str,
		  	dt_added : datetime,
		  	config : object | Configuration,
			dt_started : datetime | None = None,
		  	status : RunQueueItemStatus = RunQueueItemStatus.Queued,
		  	dt_done : typing.Union[datetime, None] = None,
		  	exit_code : typing.Union[int, None] = None,
		  	stderr : str = ""
		) -> None:
		self.item_id = item_id
		self.name = name
		self.dt_added = dt_added
		self.config = config
		self.dt_started = dt_started
		self.status = status
		self.dt_done = dt_done
		self.exit_code = exit_code
		self.stderr = stderr

	item_id: int
	name: str #Descriptive name used to display in the queue
	dt_added: datetime
	config: object
	status: RunQueueItemStatus = RunQueueItemStatus.Queued

	# done : bool = False
	dt_done : typing.Union[datetime, None] = None #When done running/cancelled/stopped etc.
	dt_started : typing.Union[datetime, None] = None #When running started
	exit_code : typing.Union[int, None] = None
	stderr : str = ""

	@staticmethod
	def copy_to(source : "RunQueueItem", target : "RunQueueItem"):
		""" Copy the data from the source item to the target item """
		target.item_id = source.item_id
		target.name = source.name
		target.dt_added = source.dt_added
		target.config = source.config
		target.dt_started = source.dt_started
		target.status = source.status
		target.dt_done = source.dt_done
		target.exit_code = source.exit_code
		target.stderr = source.stderr


	def get_copy(self):
		"""
		Get a copy of the object
		"""
		return RunQueueItem(
			item_id=self.item_id,
			name=self.name,
			dt_added=self.dt_added,
			config=copy(self.config),
			dt_started=self.dt_started,
			status=self.status,
			dt_done=self.dt_done,
			exit_code=self.exit_code,
			stderr=self.stderr
		)

class RunQueueItemActions(Enum):
	"""Describe the actions which can be performed on a run queue item"""
	DELETE = 0
	STOP = 1 #A more forceful "cancel" which will also kill the process if it is running
	MOVEUP = 2
	MOVEDOWN = 3
	MOVETOP = 4
	# Requeue = 5 #Re-put #TODO: reput the item in queue? Or reput copy of item in queue?
	CANCEL = 6 #Maybe name as "pause"
	START = 7 #Force-start a queued item (TODO: maybe also cancelled items?)

class CustomManager(managers.SyncManager):
	"""
	Source:
	https://docs.python.org/3/library/multiprocessing.html#multiprocessing.managers.SyncManager
	allows the default object to be shared between processes
	"""

class CustomProxy(managers.NamespaceProxy): #type:ignore
	"""
	#NOTE: Normally, the proxy used by BaseManager only exposes methods from the object it is referring to.
	This class exposes all attributes - but not the methods"""
	_exposed_ = ('__getattribute__', '__setattr__', '__delattr__')

CustomManager.register(RunQueueItem.__name__, RunQueueItem, CustomProxy)
	#Register the class so that it can be shared
	#  between processes NOTE: making changes in the config itself will not be propagated - only if the whole object is
	# replaced (https://docs.python.org/3/library/multiprocessing.html#multiprocessing-proxy-objects)


class CommandlineQueueEmitter():
	"""
	A class that keeps track of a command-line-queue and emits a signal each time an item is added to the queue.
	Enables the use of multiple threads, while still logging to the UI without polling files.
	"""
	# commandLineOutput = QtCore.Signal(int, str, str, datetime, int, str) #id, name, output_path, dt, filepos, new_msg
	commandLineOutput = PySignal.ClassSignal() #int, str, str, datetime, int, str

	def __init__(self, monitored_queue : typing.Union[queue.Queue, multiprocess.queues.Queue]) -> None:
		super().__init__()
		self._monitored_queue = monitored_queue
		self.stop_flag = False

	def run(self):
		"""Worker function that continuously waits for new items in the queue and emits a signal when a new item is
		found, if stop_flag is set, the thread will stop after the next timeout (<1s)
		"""
		while not self.stop_flag: #Continuously wait for new items in the queue
			try:
				args = self._monitored_queue.get(block=True, timeout=0.5) #Timeout every 0.5 seconds to check if we should stop
				self.commandLineOutput.emit(*args)
			except queue.Empty:
				pass
			except BrokenPipeError:
				log.warning("Broken pipe error in CommandlineQueueEmitter - if this occured while app was closing, you can "
	      				"ignore this warning. Queue-emitter will now stop monitoring the command-line output queue.")
				self.stop_flag = True


class RunQueue():
	"""
	A class in which we can queue configurations to run tasks. The Configurations are ran in separate processes.
	We can turn on automatic processing, which will automatically start processing items in the queue if the number
	of processes allows for it.

	TODO: We should also be able to manually start processing items in the queue.
	"""


	######################### OLD PYSIDE 6 VERSION #########################
	# queueChanged = QtCore.Signal(object) #Emits a snapshot of the current queue when the queue changes (list of ints)
	# # runListChanged = QtCore.Signal(object) #Emits a snapshot of the all_list when the all_list changes
	# # 		# type is (typing.Dict[int, RunQueueItem])
	# allItemsDictInsertion = QtCore.Signal(list, object) #Emits a list of id(s) and the new item-dict when new item(s) are
	# 	# inserted we can utilize this in qt-models to update (Tree/Table) models by just inserting rows instead of resetting
	# allItemsDictRemoval = QtCore.Signal(list, object) #Emits a list of id(s) and the new all_dict when item(s) is/are
	# 	# removed we can utilize this in qt-models to update (Tree/Table) models by just removing rows instead of resetting

	# itemDataChanged = QtCore.Signal(int, object) #Emits an id with the new RunQueueItem when a single item in the
	# 		# all_dict has been changed

	# queueRunStateChanged = QtCore.Signal(bool) #True if the queue is running, False if it is no longer running
	# autoProcessingStateChanged = QtCore.Signal(bool) #Emits True if autoprocessing is enabled, False if it is disabled

	# #TODO: implement a thread that keeps watch of a separate thread in which a queue with log-updates is processed
	# newCommandLineOutput = QtCore.Signal(int, str, str, datetime, int, str) #id, name, output_path, dt, filepos, new_msg
	# currentlyRunningIdsChanged = QtCore.Signal(object) #Emits a list of ids that are currently running
	# resetTriggered = QtCore.Signal() #Emitted when the queue is reset (indicates that all models should be reset)
	####################################################################################
	

	############################New version without pyside6 dependency#######################
	queueChanged = PySignal.ClassSignal() #Emits a snapshot of the current queue when the queue changes (list of ints)
	# runListChanged = PySignal.ClassSignal(object) #Emits a snapshot of the all_list when the all_list changes
	# 		# type is (typing.Dict[int, RunQueueItem])
	allItemsDictInsertion = PySignal.ClassSignal() #Emits a list of id(s) and the new item-dict when new item(s) are
		# inserted we can utilize this in qt-models to update (Tree/Table) models by just inserting rows instead of resetting
	allItemsDictRemoval = PySignal.ClassSignal() #Emits a list of id(s) and the new all_dict when item(s) is/are
		# removed we can utilize this in qt-models to update (Tree/Table) models by just removing rows instead of resetting

	itemDataChanged = PySignal.ClassSignal() #Emits an id with the new RunQueueItem when a single item in the
			# all_dict has been changed

	queueRunStateChanged = PySignal.ClassSignal() #True if the queue is running, False if it is no longer running
	autoProcessingStateChanged = PySignal.ClassSignal() #Emits True if autoprocessing is enabled, False if it is disabled

	#TODO: implement a thread that keeps watch of a separate thread in which a queue with log-updates is processed
	newCommandLineOutput = PySignal.ClassSignal() #int, str, str, datetime, int, str = id, name, output_path, dt, filepos, new_msg
	currentlyRunningIdsChanged = PySignal.ClassSignal() #Emits a list of ids that are currently running
	resetTriggered = PySignal.ClassSignal() #Emitted when the queue is reset (indicates that all models should be reset)

	def __init__(self,
	      		target_function : typing.Callable[[Configuration], typing.Any],
				n_processes : int = 1,
				log_location : str= "",
				log_location_make_dirs : bool = True
			) -> None:
		"""_summary_

		Args:
			target_function (typing.Callable[[Configuration], typing.Any]): The function which is passed the configuration
				and which is run in a separate process.

			n_processes (int, optional): How many processes to use for processing the queue. Defaults to 1.

			log_location (str, optional): Stdout/stderr of each subprocess is logged to a file for ease-of-access from
				the main process/ui if no path is given, we try to find the temp-path (using tempfile.TemporaryFile()).
				Defaults to None.

			log_location_make_dirs (bool, optional): Whether to create the log location if it does not exist. Defaults to True.
		"""

		super().__init__()


		self._target_function = target_function
		self._stop_autoprocessing_flag = multiprocess.Event()
		self._stop_autoprocessing_flag.clear()

		self._log_location = log_location
		if self._log_location is None or self._log_location == "": #If no
				#log location is given, use the tempdir (os-specific)
			self._log_location = tempfile.gettempdir()
			#Create sub folder Configurun in tempdir so we have everything in one place
			self._log_location = os.path.join(self._log_location, "configurun")
			if not os.path.exists(self._log_location):
				os.mkdir(self._log_location)
		elif not os.path.exists(self._log_location):
			if log_location_make_dirs:
				os.makedirs(self._log_location)
			else:
				raise ValueError(f"Passed log location {self._log_location} does not exist - either create it before "
		     		"running or set it to None to use the temp-directory, or enable log_location_make_dirs to create "
					"the directory")

		log.info(f"Runqueue log location: {self._log_location}")

		self._manager = CustomManager() #To share data between processes
		self._manager.start() #Start the manager

		# self._queue : queue.Queue = queue.Queue()
		self._queue : typing.List[int] = self._manager.list() # type: ignore #Consists of a queue of id's which can be
			# used to retrieve the configuration from the all_dict - can't be an actual queue because we want to be able
			# to remove/move items in the queue
		self._queue_mutex = multiprocess.Lock() #Lock for the queue

		self._all_items_dict : typing.Dict[int, RunQueueItem] = self._manager.dict() #type: ignore #Contains all items
			# that have ever been added minus the removed ones.
			# TODO: Make this an ordered dict to make clear that new items are appended to the end, this is not
			# really necceasry for this class itself, but makes it easier to use in qt-models (tablemodels/treeviews)
			# because of a more predictable (next) order after item change/insertion/removal
		self._all_items_dict_mutex = multiprocess.Lock()


		self._autoprocessing_enabled = False #Whether to automatically start processing items in the queue
		self._running_processes : typing.Dict[int, multiprocess.Process] = {} #ID -> process
		self._running_processes_mutex = multiprocess.Lock()

		self._n_processes = n_processes
		self._cur_id = 0 #Start at 0

		self._queue_processor_thread : threading.Thread | None = None #Thread that processes the queue
		self._queue_signal_updater_thread : threading.Thread | None = None #Thread that keeps the queue updated

		self._cmd_id_name_path_dict : typing.Dict[int, typing.Tuple[str, str]] = self._manager.dict() # type: ignore
			# Dictionary that keeps track of the output file for each process
		self._cmd_id_name_path_dict_mutex = multiprocess.Lock() #Lock for the cmd_id_path_dict
		#Queue that keeps track of outputs to the command line for each running process.
		self._command_line_output_queue = self._manager.Queue()



		#OLD version with pyside6 dependency
		# self.queue_emitter = CommandlineQueueEmitter(self._command_line_output_queue)
		# self.queue_emitter_thread = QtCore.QThread()
		# self.queue_emitter.moveToThread(self.queue_emitter_thread)
		# self.queue_emitter_thread.started.connect(self.queue_emitter.run)
		# self.queue_emitter_thread.start()
		# self.queue_emitter.commandLineOutput.connect(self.newCommandLineOutput)


		#New version without pyside6 dependency
		self.queue_emitter = CommandlineQueueEmitter(self._command_line_output_queue)
		self.queue_emitter_thread = threading.Thread(target=self.queue_emitter.run)
		self.queue_emitter.commandLineOutput.connect(lambda *args: self.newCommandLineOutput.emit(*args))
		# self.queue_emitter.commandLineOutput.connect(
		# 	lambda *args: print(f"Command line changed{', '.join([str(i) for i in args])}"))
		self.queue_emitter_thread.start()


	def handle_command_line_output(self, item_id : int, name , output_path, dt , filepos, new_msg):
		self.newCommandLineOutput.emit(item_id, name, output_path, dt, filepos, new_msg)

	def stop_command_line_queue_emitter(self):
		"""
		Stop the queue-emitter thread - should be called when terminating the queue
		TODO: create and explicit cleanup function
		"""
		self.queue_emitter.stop_flag = True
		#Pyside6-version
		# self.queue_emitter_thread.quit()
		# self.queue_emitter_thread.wait()

		#Non-pyside6-version
		self.queue_emitter_thread.join()


	@staticmethod
	def set_workspace_lock(workspace_path : str, lock_filename = WORKSPACE_LOCK_FILE_NAME):
		"""
		Lock the workspace by creating a lockfile in the workspace directory.

		Raises:
			WorkspaceInUseException: If the workspace is already locked/in use
		"""
		lock_filepath = os.path.join(workspace_path, lock_filename)
		if os.path.exists(lock_filepath):
			raise WorkspaceInUseException(f"Could not lock workspace at {workspace_path}, lockfile already exists at {lock_filepath}.")
		with open(lock_filepath, "w", encoding='utf-8') as lock_file:
			lock_file.write("This file is used to indicate that the workspace at this path is in use by another instance of "
				"a configurun-app. Please only remove this file if the app was closed improperly and this file remained.")
	
	@staticmethod
	def release_workspace_lock(workspace_path : str, lock_filename = WORKSPACE_LOCK_FILE_NAME):
		"""
		Release the workspace lock by removing the lockfile from the workspace directory. If file does not exist, nothing
		is done.

		Raises:
			WorkspaceInUseException: If the workspace is already locked/in use
		"""
		lock_filepath = os.path.join(workspace_path, lock_filename)
		os.remove(lock_filepath)

	def get_command_line_output(self, item_id : int, fseek_end : int, max_bytes : int) -> typing.Tuple[str, datetime]:
		"""
		Get the command line output for a given id.

		args:
			id (int): the id of the item to get the command line output for
			fseek_end (int): the position in the file to end reading at, -1 is EOF
			max_bytes (int): the maximum number of bytes to read, -1 for all bytes

		return (str, datetime.datetime): the command line output for the given id and the datetime at which the output
			was last updated
		"""
		last_edit_dt = datetime(1970,1,1)
		with self._cmd_id_name_path_dict_mutex:
			if item_id not in self._cmd_id_name_path_dict or not os.path.exists(self._cmd_id_name_path_dict[item_id][1]):
				return ("", last_edit_dt)
			#TODO: if file doesn't exist - return string to indicate file is missing?

			last_edit_dt = datetime.fromtimestamp(os.path.getmtime(self._cmd_id_name_path_dict[item_id][1]))
			filepath = self._cmd_id_name_path_dict[item_id][1]

		with open(filepath, "r", encoding="utf-8") as read_file:
			if fseek_end == -1 and max_bytes == -1: #If just reading the whole file
				return read_file.read(), last_edit_dt
			elif fseek_end == -1: #If reading to the end of the file
				#Start reading x bytes from the end of the file
				read_file.seek(0, os.SEEK_END)
				file_size = read_file.tell()
				read_file.seek(max(0, file_size - max_bytes), os.SEEK_SET)
			if max_bytes == -1:
				return read_file.read(), last_edit_dt

			return read_file.read(max_bytes), last_edit_dt

	def get_command_line_info_list(self) -> typing.Dict[int, typing.Tuple[str, str, int, bool]]:
		"""
		Get a list of the info of all command lines outputs from items that have been ran or are running.

		returns:
			dict[int, tuple[str, str, int, bool]]: a dictionary with the id as key and a tuple with the:
			 - name of the item
			 - path to the output file
			 - length of the output file
			 - Whether the id is currently running (bool)
		"""
		with self._cmd_id_name_path_dict_mutex:
			# return copy(dict(self._cmd_id_name_path_dict)) #TODO: copy probably unnecesary after converting
			cur_dict = dict(self._cmd_id_name_path_dict)

		new_dict = {}

		for key, val in cur_dict.items():
			if os.path.exists(val[1]) and os.path.isfile(val[1]):
				length = os.path.getsize(val[1])
			else:
				length = -1 #Indicate file not found

			currently_running = False
			with self._running_processes_mutex:
				currently_running = key in self._running_processes #If running

			new_dict[key] = (val[0], val[1], length, currently_running)
		return new_dict

	@staticmethod
	def get_queue_contents_dict_from_file(path : str, allow_load_running_items : bool = False):
		"""
		Load all data neccesary to create queue from a file.
		Settings such as _n_processes are not loaded.
		TODO: maybe enforce JSON?

		Args:
			path (str): the path to the file to load the queue from
			allow_load_running_items (bool, optional): Whether to allow loading a queue that contains running items.
				Defaults to False.

		Returns:
			dict[str, typing.Any]: a dictionary containing the contents of the queue to load.
				Actual queue can then be loaded using load_file_contents_dict.
		"""
		if not os.path.exists(path):
			raise OSError(f"Could not load queue-data from path {path}, path does not exist.")

		with open(path, "rb") as load_file:
			contents_dict = dill.load(load_file) #TODO: maybe separate thread? block until done. Probably small files though.
		assert isinstance(contents_dict, dict), "Could not load queue-data from file, file does not contain a RunQueue-dict."
		if not allow_load_running_items and contents_dict.get("had_running_items", False):
			raise RunQueueHasRunningItemsException(f"Could not load queue-data from file {path}, file contains running "
				" items and allow_load_running_items is set to False.")

		neccesary_keys = "all_items_dict", "queue_copy", "cur_id", "cmd_id_name_path_dict"
		for key in neccesary_keys:
			if key not in contents_dict:
				raise KeyError(f"Could not load a queue from file {path}, file does not contain (required) key: {key}.")

		return contents_dict


	def load_queue_contents_dict(self,
				contents_dict : typing.Dict[str, typing.Any]
			):
		"""
		Resets the queue contents using the passed arguments. This is a convenience function that can be used to load
		the queue from a file.

		Args:
			contents_dict (typing.Dict[str, typing.Any]): a dictionary containing the contents of the queue to load.
				We can get this dict from an instance using get_queue_contents_dict().
				We can get this dict from a file using get_queue_contents_dict_from_file().
				We can set the dict to be loaded by an instance using this function.
		"""
		log.info("Now resetting queue using passed contents dict")
		self.stop_autoprocessing() #Stop autoprocessing when loading (though the locks should prevent any issues)

		with self._all_items_dict_mutex, self._queue_mutex:
			if self._get_running_configuration_count_nolocks() > 0:
				raise RuntimeError("Could not load queue from dictionary, queue contains running items. Please make sure"
		       		" no configurations are running when loading queue data.")

			#NOTE: We have to make sure that the objects are managed by the manager, otherwise changes will not propagate
			# between processes
			self._all_items_dict = self._manager.dict(contents_dict["all_items_dict"])

			for key, run_queue_item in self._all_items_dict.items():
				#For each run_queue_item, create a managed instance of the RunQueueItem class
				#TODO: is there a neater way to construct the managed RunQueueItem without manually copying all the data?
				item_copy = copy(run_queue_item)
				self._all_items_dict[key] = self._manager.RunQueueItem(None, None, None, None, None)
				RunQueueItem.copy_to(item_copy, self._all_items_dict[key]) #Copy over the item data

			self._queue = self._manager.list(contents_dict["queue_copy"])
			self._cur_id = contents_dict["cur_id"]
			self._cmd_id_name_path_dict = self._manager.dict(contents_dict["cmd_id_name_path_dict"]) #Load path-locations
			#of the cmd outputs
			#TODO: clear commandline output queue as well
			#TODO: make cmd output relative to the workspace folder? That way we can copy between machines.
		self.resetTriggered.emit() #Emit signal to indicate that the queue has been reset

	def get_queue_contents_dict(self, save_running_as_stopped : bool = False):
		"""
		Save the queue to a file - using pickle as to make things easy no matter what type of settings object is used.
		Runqueue-settings such as _n_processes are not saved

		NOTE: This function will not save the queue if there are still items running, unless save_running_as_stopped is
		set to True. If this is the case, items that were running will be saved as being stopped. stderr will be set to
		a msg indicating that the process was running when the queue was saved and that it might have actually finished.

		Args:
			path (str): the path to the file to save the queue to
		"""
		# with open(path, "wb") as file:
		with self._all_items_dict_mutex, self._queue_mutex: #Make sure queue and all_dict are consistent with one another
			all_items_dict_copy = self._get_all_items_dict_snapshot_copy_nolocks()
			queue_copy = self._get_queue_snapshot_copy_no_locks()
			running_ids_count = self._get_running_configuration_count_nolocks()

		had_running_items = False
		if running_ids_count > 0:
			had_running_items = True
			if not save_running_as_stopped:
				raise RunQueueHasRunningItemsException("Could not create a copy of runqueue-data, the queue contains "
					" configurations that are currently running and save_running_as_stopped is set to False.")


		for runqueue_item in all_items_dict_copy.values():
			if runqueue_item.status == RunQueueItemStatus.Running:
				runqueue_item.status = RunQueueItemStatus.Stopped
				runqueue_item.stderr = "Process was running when queue was saved, so it was saved as stopped."
				runqueue_item.dt_done = datetime.now()

		return {
				"all_items_dict" : all_items_dict_copy,
				"queue_copy" : queue_copy,
				"cur_id" : self._cur_id, #Make sure we don't start overwriting items
				"cmd_id_name_path_dict": copy(dict(self._cmd_id_name_path_dict)), #Save the file-locations of the
					# command line outputs
					#NOTE: we have to copy the dict, otherwise we get errors when loading it again since it's a managed
					# object which can't survive between app-runs (results in a FileNotFoundError when trying to load)
				"had_running_items" : had_running_items #So we can indicate when loading that there were running items
		}

	def set_item_config(self, item_id : int, new_config : Configuration):
		"""
		Set the configuration of an item in the queue.

		Args:
			item_id (int): the id of the item to set the configuration for
			new_config (Configuration): the new configuration to set

		Raises:
			KeyError: if the item_id config cannot be set. This can only be done when the item is cancelled or
				when it's in-queue. Already-finished/stopped items cannot be changed to make sure the history is
				accurate
		"""
		with self._all_items_dict_mutex:
			if self._all_items_dict[item_id].status in [
								RunQueueItemStatus.Queued,
								RunQueueItemStatus.Cancelled
						   ]:

				self._all_items_dict[item_id].config = new_config
				self.itemDataChanged.emit(item_id, self._all_items_dict[item_id].get_copy()) #NOTE: for network, we have to un-proxy the object
			else:
				raise ConfigurationIsFirmException(
					f"Could not set configuration for item with id {item_id}. \n\nCannot change the item "
		   			f"configuration when it is in state: {self._all_items_dict[item_id].status}"
				)


	def get_item_config(self, item_id : int):
		"""
		Get the configuration of an item in the queue.

		Args:
			item_id (int): the id of the item to get the configuration for
		Raises:
			KeyError: if the item_id config is not a know id (not in all_items_dict)
		"""
		with self._all_items_dict_mutex:
			if item_id not in self._all_items_dict:
				raise KeyError(f"Could not get configuration for item with id {item_id}, id not in queue.")
			return self._all_items_dict[item_id].config


	@staticmethod
	def get_actions_from_status(status : RunQueueItemStatus) -> typing.List[RunQueueItemActions]:
		"""
		Retrieve the possible actions (RunQueueItemActions) that can be performed on an item with status

		"""
		if status == RunQueueItemStatus.Stopped \
					or status == RunQueueItemStatus.Finished:
			actions = [
				RunQueueItemActions.DELETE #TODO: Add requeue
			]
		elif status == RunQueueItemStatus.Cancelled:
			actions = [
				RunQueueItemActions.DELETE, #TODO: Add requeue
				RunQueueItemActions.START
			]
		elif status == RunQueueItemStatus.Failed:
			actions = [
				RunQueueItemActions.DELETE #TODO: Add retry
			]
		elif status == RunQueueItemStatus.Running:
			actions = [
				RunQueueItemActions.STOP
			]
		elif status == RunQueueItemStatus.Queued:
			actions = [
				RunQueueItemActions.DELETE,
				RunQueueItemActions.CANCEL, #Remove from queue (="cancel")
				RunQueueItemActions.MOVEUP,
				RunQueueItemActions.MOVEDOWN,
				RunQueueItemActions.MOVETOP,
				RunQueueItemActions.START
			]
		else:
			raise KeyError(f"Could not get actions due to unknown status:"
					f"{str(status)}"
				)
		return actions


	def get_actions_for_id(self, item_id : int | None) -> typing.List[RunQueueItemActions]:
		"""
		Get a list of actions that can be performed on the item with id

		Args:
			id (int): the id of the item to get the actions for
		"""
		actions = []
		if item_id is None:
			return actions
		if item_id in self._queue:
			actions = [

			]
		elif item_id in self._all_items_dict:
			#Lock queue when removing
			with self._all_items_dict_mutex, self._queue_mutex:
				actions = RunQueue.get_actions_from_status(self._all_items_dict[item_id].status)



		return actions

	def is_autoprocessing_enabled(self):
		"""Whether autoprocessing is enabled"""
		return self._autoprocessing_enabled

	def do_action_for_id(self, item_id : int, action : RunQueueItemActions):
		"""
		Do an action for a given id

		Args:
			id (int): the id of the item to do the action for
			action (QueueItemActions): the action to perform on the item
		"""
		log.info("Inside runqueue.do_action_for_id")
		if action == RunQueueItemActions.DELETE:
			self.delete_id(item_id)
		elif action == RunQueueItemActions.STOP:
			self.force_stop_id(item_id)
		elif action == RunQueueItemActions.CANCEL:
			self.cancel_id(item_id)
		elif action == RunQueueItemActions.MOVEUP:
			self.rel_move_in_queue(item_id, -1)
		elif action == RunQueueItemActions.MOVEDOWN:
			self.rel_move_in_queue(item_id, 1)
		elif action == RunQueueItemActions.MOVETOP:
			self._move_id_in_queue(item_id, 0)
		elif action == RunQueueItemActions.START:
			self.start_running_id(item_id)
		else:
			log.warning(f"Unknown action: {str(action)}")
			raise NotImplementedError(f"Unknown action: {str(action)}")




	def cancel_id(self, item_id):
		"""
		Cancel by id -> Remove a configuration from the queue (by id), but don't delete it

		Args:
			id (int): the id of the item to cancel
		"""
		#Lock queue when removing
		with self._all_items_dict_mutex, self._queue_mutex:
			assert self._all_items_dict[item_id].status == RunQueueItemStatus.Queued, ("Can only cancel queued items but item "
				"is currently in status: " + str(self._all_items_dict[item_id].status))
			if item_id not in self._queue: #Only do something if the id is in the queue
				raise KeyError(f"Could not cancel {item_id}, ID not in queue.")
			self._all_items_dict[item_id].status = RunQueueItemStatus.Cancelled
			self._queue.remove(item_id)
			queue_snapshot = self._get_queue_snapshot_copy_no_locks()
			item_copy = self._all_items_dict[item_id].get_copy()

		self.queueChanged.emit(queue_snapshot)
		self.itemDataChanged.emit(item_id, item_copy)
		return



	def _move_id_in_queue(self, item_id : int, new_queue_pos : int, use_locks=True):
		"""
		Move a configuration in the queue.

		Args:
			id (int): the id of the item to move
			new_queue_pos (int): the new position in the queue
			use_locks (bool, optional): whether to use locks when moving the item. Defaults to True. Is used internally
				when items are already locked
		"""
		#Lock queue when removing (unless we are already holding the locks)
		if use_locks:
			self._all_items_dict_mutex.acquire()
			self._queue_mutex.acquire()

		if item_id in self._queue:
			self._queue.remove(item_id)
			self._queue.insert(new_queue_pos, item_id)
		else: #Throw exception
			if use_locks:
				self._all_items_dict_mutex.release()
				self._queue_mutex.release()
			raise KeyError(f"Could not move item with id {item_id}, ID not in queue.")

		self.queueChanged.emit(self._get_queue_snapshot_copy_no_locks())
		if use_locks:
			self._all_items_dict_mutex.release()
			self._queue_mutex.release()



	def rel_move_in_queue(self, item_id : int, rel_move : int):
		"""Relative move in the queue, -1 = move up, 1 = move down. 0 is the top of the queue.

		Args:
			id (int): the id of the item to move
			rel_move (int): the mount to move the item. Negative values move the item up in the queue, positive values
			  move it down
		"""
		log.info(f"Moving item with id {item_id} by {rel_move} positions")
		with self._all_items_dict_mutex, self._queue_mutex:
			if item_id in self._queue:
				cur_pos = self._queue.index(item_id)
				new_pos = min(len(self._queue), max(0, cur_pos + rel_move))
				self._move_id_in_queue(item_id, new_pos, use_locks=False) #We already acquired the locks
				log.debug(f"Moved item with id {item_id} from position {cur_pos} to {new_pos}")
			else:
				self._all_items_dict_mutex.release()
				self._queue_mutex.release()
				raise KeyError(f"Could not move item with id {item_id}, ID not in queue.")





	def delete_id(self, item_id : int):
		"""
		Remove a configuration from the queue and list
		args:
			id (int): the id of the item to remove
		"""
		#Lock queue when removing
		with self._all_items_dict_mutex, self._queue_mutex:
			if self._all_items_dict[item_id].status == RunQueueItemStatus.Running:
				self._all_items_dict_mutex.release()
				self._queue_mutex.release()
				raise NotImplementedError(f"Could not delete item with id {item_id}, item is running.") #TODO: Add force del?

			if item_id in self._queue:
				self._queue.remove(item_id)

			#Also remove from all_dict -> the two should be in sync
			del self._all_items_dict[item_id]

			run_list_snapshot = self._get_all_items_dict_snapshot_copy_nolocks()
			queue_snapshot = self._get_queue_snapshot_copy_no_locks()

		self.queueChanged.emit(queue_snapshot)
		self.allItemsDictRemoval.emit([item_id], run_list_snapshot)




	def get_queue_snapshot_copy(self) -> list[int]:
		"""
		Get a snapshot of the queue

		returns:
			list[int]: a copy of the queue (list of ids)
		"""
		#Lock queue when getting snapshot
		with self._queue_mutex:
			snapshot = copy(list(self._queue)) #TODO: copy might be unnecesary after converting proxylist to list
		return snapshot

	def _get_queue_snapshot_copy_no_locks(self) -> list[int]:
		snapshot = copy(self._queue)
		return list(snapshot)


	def _get_all_items_dict_snapshot_copy_nolocks(self) -> typing.Dict[int, RunQueueItem]:
		"""Get a snapshot of the all_list without locks - should only be used internally when locks are already held

		Returns:
			typing.Dict[int, RunQueueItem]: A copy of the run_list
		"""
		ret_dict = {}
		for key, value in self._all_items_dict.items(): #Since this is a nested proxy-object, call get_copy on each item
			ret_dict[key] = value.get_copy()
		return ret_dict

	def get_all_items_dict_snapshot_copy(self) -> typing.Dict[int, RunQueueItem]:
		"""
		Get a snapshot of the all_list
		"""
		with self._all_items_dict_mutex:
			snapshot = self._get_all_items_dict_snapshot_copy_nolocks()
		return snapshot


	def start_autoprocessing(self):
		"""
		Start automatically processing items in the queue
		"""
		if not self._autoprocessing_enabled:
			self._autoprocessing_enabled = True
			self.autoProcessingStateChanged.emit(True)
			self._run_queue()

	def get_n_processes(self):
		"""
		Get the number of processes this runqueue can use (at max).
		If -1, the number of processes is unlimited.
		"""
		return self._n_processes

	def set_n_processes(self, n_processes : int):
		"""
		Set the number of processes, does not affect currently running processes, but might start new ones
		if the number of running processes is below the new n_processes value and autoprocessing is enabled.

		Args:
			n_processes (int): the new number of processes to use
		"""
		self._n_processes = n_processes

	def get_running_configuration_count(self):
		"""
		Get a list of the currently running configurations while locking the queue and all_items_dict
		"""
		with self._all_items_dict_mutex, self._queue_mutex:
			return self._get_running_configuration_count_nolocks()

	def _get_running_configuration_count_nolocks(self):
		count = 0
		for queue_item in self._all_items_dict.values():
			if queue_item.status == RunQueueItemStatus.Running:
				count+=1
		return count

	# def _wait_stop(self, timeout = 10_000):
	# 	self._stopflag.set()
	# 	start_time = time.time()
	# 	for process in self._running_processes.values():
	# 		remaining_time = start_time + timeout - time.time()
	# 		process.join(timeout )

	# 	self._running = False

	def stop_autoprocessing(self):
		"""
		Stop automatically processing new items, finish processing the current items and then stop the queue.
		"""
		if self._autoprocessing_enabled:
			self._autoprocessing_enabled = False
			self._stop_autoprocessing_flag.set()
			self.autoProcessingStateChanged.emit(False)

	def force_stop_all_running(self,
				stop_msg : str = "Process was force stopped by user."
			):
		"""
		Force stop all running processes and set the configuration to cancelled
		NOTE: also stops autoprocessing

		Args:
			stop_msg (str, optional): the message to set as stderr for the stopped items. Defaults to
				"Process was force stopped by user.".
		"""
		log.info("Now force stopping all running processes...")
		self.stop_autoprocessing()

		#First terminate all processes
		with self._running_processes_mutex:
			stopped_ids = list(self._running_processes.keys())
			for item_id in stopped_ids:
				if self._running_processes[item_id].is_alive(): #Terminate if still alive
					try:
						self._running_processes[item_id].terminate() #TODO: what if just about to finish?
					except Exception as exception: #pylint: disable=broad-except
						log.warning(f"Could not terminate process with id {item_id} - "
		  					f"{type(exception).__name__}: {str(exception)}")
				del self._running_processes[item_id]

		self.currentlyRunningIdsChanged.emit([]) #Nothing should be running anymore

		with self._all_items_dict_mutex, self._queue_mutex:
			for item_id in stopped_ids:
				self._all_items_dict[item_id].status = RunQueueItemStatus.Stopped
				self._all_items_dict[item_id].dt_done = datetime.now()
				self._all_items_dict[item_id].exit_code = -1
				self._all_items_dict[item_id].stderr = stop_msg

		for item_id in stopped_ids: #Update GUI for all stopped items
			self.itemDataChanged.emit(item_id, self._all_items_dict[item_id].get_copy())



	def save_and_stop_all(self, save_path : str):
		"""
		Save the queue and then try to stop all updaters/etc.
		All running processes will be force stopped and set to a 'stopped' state.

		Args:
			save_path (str): the path to save the queue to
		"""
		self.stop_autoprocessing()

		with self._all_items_dict_mutex, self._queue_mutex: #Make sure we don't start anything new while stopping
			self.force_stop_all_running()


		contents_dict = self.get_queue_contents_dict(save_running_as_stopped=True)
		with open(save_path, "wb") as save_file:
			dill.dump(contents_dict, save_file)




	def force_stop_id(self, item_id : int):
		"""
		Force stop a single running process and set the configuration to cancelled

		Args:
			item_id (int): the id of the item to stop

		Raises:
			KeyError: if the item is not running, or if item not in queue
		"""
		with self._running_processes_mutex:
			if item_id not in self._running_processes:
				raise KeyError(f"Could not force stop item with id {item_id}, item is not running.")

			self._running_processes[item_id].terminate()
			del self._running_processes[item_id]
			running_ids = list(self._running_processes.keys())

		self.currentlyRunningIdsChanged.emit(running_ids) #TODO: move inside lock?

		with self._all_items_dict_mutex, self._queue_mutex:
			self._all_items_dict[item_id].status = RunQueueItemStatus.Stopped
			self._all_items_dict[item_id].dt_done = datetime.now()
			self._all_items_dict[item_id].exit_code = -1
			self._all_items_dict[item_id].stderr = "Process was force stopped by user."
			queue_snapshot = self._get_queue_snapshot_copy_no_locks()
			item_copy = self._all_items_dict[item_id].get_copy()

		self.queueChanged.emit(queue_snapshot) #TODO: move inside lock?
		self.itemDataChanged.emit(item_id, item_copy)


	def add_to_queue(self, name : str, config : Configuration):
		"""
		Add a configuration to the queue

		Args:
			name (str): What name to display in the queue
			config (typing.Any): The configuration (data) that is held by this item. Is only processed by the target
				of the run queue
		"""
		log.debug(f"Adding config with name {name} to queue.")
		new_item_id = self._cur_id
		self._cur_id += 1
		dt_added = datetime.now()
		new_item = self._manager.RunQueueItem(item_id=new_item_id,name=name, dt_added=dt_added, config=config) #pylint: disable=no-member #type: ignore
		# new_item = RunQueueItem(item_id=new_item_id,name=name, dt_added=dt_added, config=config)

		with self._all_items_dict_mutex, self._queue_mutex:
			self._queue.append(new_item_id)
			self._all_items_dict[new_item_id] = new_item
			snapshot = self._get_all_items_dict_snapshot_copy_nolocks()
			queue_snapshot = self._get_queue_snapshot_copy_no_locks()

		self.allItemsDictInsertion.emit([new_item_id], snapshot)
		self.queueChanged.emit(queue_snapshot) #Emit outside of lock

	def _run_queue(self):
		"""
		Run the queue (the producer), removes items from the queue and puts them in the currently-processing-queue
		(at which point the consumer will pick them up, and editing will no longer be possible)
		"""
		# self._stopflag = multiprocessing.Event()
		self._stop_autoprocessing_flag.clear()

		if self._queue_processor_thread is None or not self._queue_processor_thread.is_alive():
			self._queue_processor_thread = threading.Thread(target=self._run_queue_item_processor)
		self._queue_processor_thread.start()


		#Create thread and run _run_queue_item_producer if it doesn't exist yet
		if self._queue_signal_updater_thread is None or not self._queue_signal_updater_thread.is_alive():
			self._queue_signal_updater_thread = threading.Thread(target=self._run_queue_item_updater)
			self._queue_signal_updater_thread.start()



	def _run_queue_item_updater(self):
		#While we are not stopping, or there are still processes running, keep updating
		while not self._stop_autoprocessing_flag.is_set() or len(self._running_processes) > 0:
			# self.runListChanged.emit(self.get_run_list_snapshot_copy()) #For now, just update the whole list every x s.
			for item_id in copy(list(self._running_processes.keys())): #TODO: not very neat... but updates items that are running
					# since they are running in another thread, we probably want to create a queue with item-updates
					# and then update the items in the queue in this thread by checking if the item is in the queue
					# this way, we don't update unneccesarily
				item = self._all_items_dict.get(item_id, None)
				if item: #Makes sure that item is not deleted just when we are updating it
					self.itemDataChanged.emit(item_id, item.get_copy())
			# for item_id in self.get_run_list_snapshot_copy().keys():
				# self.itemDataChanged.emit(item_id, self._all_items_dict[item_id])

			self.queueChanged.emit(self.get_queue_snapshot_copy())
			time.sleep(1)




	@staticmethod
	def _process_queue_item(
				target_function : typing.Callable[[Configuration], typing.Any],
				queue_item_id : int,
				queue_item_name : str,
				command_line_output_path : str,
				command_line_output_queue : multiprocess.queues.Queue,
				all_dict : typing.Dict[int, RunQueueItem],
				id_queue : typing.List[int],
				all_dict_mutex : threading.Lock,
				queue_mutex : threading.Lock
			): #TODO: add log_changed queue
		"""
		Process a single queue item
		"""
		log.info(f"Now processing queue item {queue_item_id}")

		#Output stdou/stderr to file
		# file = open(std_path, "a", buffering=1)

		# TODO: might be more neat to log using the logging module instead - but
		root = logging.getLogger()
		root.setLevel(logging.DEBUG)
		handler = FileAndQueueHandler(
				item_id=queue_item_id,
				item_name=queue_item_name,
				log_queue=command_line_output_queue,
				log_filename=command_line_output_path,
				mode="a",
				encoding="utf-8"
			)
		handler.setLevel(logging.DEBUG)
		formatter = logging.Formatter("[{pathname:>90s}:{lineno:<4}] {asctime}  {levelname:<7s}   {message}", style='{')
		handler.setFormatter(formatter)
		root.addHandler(handler)

		sys.stdout = LoggerWriter(log.info)
		sys.stderr = LoggerWriter(log.error)


		with all_dict_mutex, queue_mutex:
			queue_item = all_dict[queue_item_id]
			queue_item.status = RunQueueItemStatus.Running #Should no longer be editable -> probably already done in
			 #_run_queue_item_processor, but just to be sure
			queue_item.dt_started = datetime.now()

		try:
			log.info(f"Started running queue item {queue_item_id} inside process {os.getpid()}")
			configuration = copy(queue_item.config) #Just to be sure we don't change the original config, make a copy
			# config.set_option_data(options_data) #NOTE: Although Options allow for a lot of types -
			# #in the runqueue, we should always expect a 'OptionsData' type
			assert isinstance(configuration, Configuration), (f"Expected runqueue item config to be of type Configuration,"
				f"but got {type(configuration)} instead.")
			target_function(configuration)
			log.info("Done running target function.")

		except Exception as exception: # pylint: disable=broad-exception-caught #Catch all exceptions in this runqueue
			msg = f"{type(exception).__name__}:{exception}"
			log.error(msg)

			#use log.error to print traceback
			log.error(traceback.format_exc())

			if queue_item: #If queue item is set -> also report item as error
				with all_dict_mutex:
					queue_item.status = RunQueueItemStatus.Failed
					queue_item.dt_done = datetime.now()
					queue_item.exit_code = -1 #error exit code
					queue_item.stderr = msg
			# raise e #TODO: create
			log.error(f"Encountered error while running queue item {queue_item_id} inside process {os.getpid()}..."
	     		f"Terminating process.")
			return


		with all_dict_mutex, queue_mutex:
			queue_item.status = RunQueueItemStatus.Finished
			queue_item.dt_done = datetime.now()
			queue_item.exit_code = 0 #normal exit code


		log.info(f"Processed item with id {queue_item.item_id} and name {queue_item.name}, "
	   		f"queue size is now {len(id_queue)}.")

	def start_running_id(self, item_id : int):
		"""Force-start running an item that is in the queue

		NOTE: this does not consider n_threads and will instead just start the item regardless of how many threads are
		currently available
		"""

		with self._all_items_dict_mutex, self._queue_mutex:
			if item_id in self._queue: #Remove from queue if it is in the queue
				self._queue.remove(item_id)

			queue_item_name = self._all_items_dict[item_id].name
			self._all_items_dict[item_id].status = RunQueueItemStatus.Running #Should no longer be editable
			item_copy = self._all_items_dict[item_id].get_copy()
		self.itemDataChanged.emit(item_id, item_copy)

		log.debug(f"Attempting to start process-function for item with id {item_id}")
		#TODO: run the actual item -> maybe wrap in a try/except to catch errors. Make sure no locks are
		#  held when running the item


		cur_logfile_loc = os.path.join(self._log_location, f"{item_id}_{queue_item_name}")
		number = 1
		extra_string = ""
		while os.path.exists(f"{cur_logfile_loc}{extra_string}.out"): #If file exists, add number
			#to the end NOTE: otherwise, if a run fails and is restarted, the old log will be overwritten
			extra_string = f"_{number}"
			number += 1
		cur_logfile_loc = f"{cur_logfile_loc}{extra_string}.out"
		#Create file NOTE: we need to create the file before emitting the signal, otherwise we can't start
		#  "watching" the file
		open(cur_logfile_loc, "w", encoding="utf-8").close()
		log.debug(f"Will log to file: {cur_logfile_loc}")
		with self._cmd_id_name_path_dict_mutex:
			self._cmd_id_name_path_dict[item_id] = (queue_item_name, cur_logfile_loc)

		# self.newRunConsoleOutputPath.emit(queue_item_id, queue_item_name, cur_logfile_loc)

		assert(item_id not in self._running_processes), "Queue item id already in running processes"
		with self._running_processes_mutex:
			self._running_processes[item_id] = multiprocess.Process(
					target=RunQueue._process_queue_item,
					args = (
						self._target_function,
						item_id,
						queue_item_name,
						cur_logfile_loc,
						self._command_line_output_queue,
						self._all_items_dict,
						self._queue,
						self._all_items_dict_mutex,
						self._queue_mutex,
					)
				)

			self._running_processes[item_id].start() #Start processing the item
			running_ids = list(self._running_processes.keys())

		self.currentlyRunningIdsChanged.emit(running_ids)
		log.info(f"Started processing item with id {item_id} and name {queue_item_name}, "
				f"queue is now {self._queue}, all_dict is of size {len(self._all_items_dict)}")


	def _run_queue_item_processor(self):
		"""
		To be ran in a separate thread.
		1. Continuously checks the queue for items to process, then calls the _run_queue_item function to process the item
			If there is process-space (n_processes), it will start a new process to process the item
			Checks the processes for completion and updates the queue accordingly

		2. Continuously checks the processes for completion and updates the queue accordingly
		"""
		log.debug("Now running queue item processor")
		while not self._stop_autoprocessing_flag.is_set():
			try:
				if len(self._queue) > 0\
						and (len(self._running_processes) < self._n_processes \
							or self._n_processes == -1): #If there are items in the
							# queue and there is process-space, we can start a new process
					with self._all_items_dict_mutex, self._queue_mutex: #Get the first item in the queue
						queue_item_id = self._queue.pop(0)
					self.start_running_id(queue_item_id)

				else: #TODO: maybe instead pass a "done-queue" to process_queue_item and have it put itself inside
					updated = False
					with self._running_processes_mutex:
						id_list = list(self._running_processes.keys())
						for cur_id in id_list:
							if not self._running_processes[cur_id].is_alive() \
									and self._running_processes[cur_id].exitcode is not None: #If the process is finished
								self._running_processes[cur_id].join()
								del self._running_processes[cur_id]
								updated = True
						running_ids = list(self._running_processes.keys())

					if not updated: #If we achieved nothing, sleep for a bit
						time.sleep(0.4)
					else:
						self.currentlyRunningIdsChanged.emit(running_ids)

			except queue.Empty:
				pass
		log.debug("Queue item processor was stopped...")




if __name__ == "__main__":
	from PySide6 import QtWidgets  # pylint: disable=ungrouped-imports
	test_app = QtWidgets.QApplication(sys.argv)
	test_queue = RunQueue(lambda *_: print("Target function has been called... Now quitting..."))
	log.info("Created a queue")
	test_app.exec()



"""
Implements the base-options from which all user-defined options should inherit
"""


import logging
import typing
from dataclasses import dataclass

from pyside6_utils.classes import \
    Serializable  # Absolute import to avoid import PySide6

log = logging.getLogger(__name__)


class InstanceMeta(type):
	"""
	Metaclass for BaseOptions, this contains an instancecheck method that allows us to check if an instance of a class
	is present in the options dict of a Configuration instance.

	This means that we can use typehinting more easily.

	NOTE: this also means that instance-checks might now wor
	"""

	def __instancecheck__(cls, __instance: typing.Any) -> bool:
		"""
		If __instance is not a Configuration instance: use the normal behaviour

		If __instance is a Configuration instance:
		Returns whether the given instance is present in the options dict, which means that we can access its attributes
		using the . operator. This is especially helpful for typehinting, e.g.:
		
		config = Configuration()
		config.options["model_options"] = SklearnModelOptions()
		assert(isinstance(config, SklearnModelOptions)) #True
		config.<sklearn_model_options_attribute> #This will work, and will also show typehinting/autofill in IDEs


		NOTE: if we add methods to the BaseOptions class, this means that it LOOKS like we can call these methods, but
		this will not work as the Configuration-wrapper only exposes the attributes of the sub-options classes, not
		their methods. TODO: maybe warn?
		"""
		#pylint: disable=import-outside-toplevel
		from configurun.configuration.configuration import Configuration #Import here to avoid circular import 
		if type(__instance) == Configuration: #pylint: disable=unidiomatic-typecheck
			return __instance.hasinstance(cls)

		return super().__instancecheck__(__instance) #Otherwise use normal behaviour


@dataclass
class BaseOptions(Serializable, object, metaclass=InstanceMeta):
# class BaseOptions(Serializable, object):
	"""
	he base-option class, all other options classes should inherit from this class
	A configuration is built by using 1 or more of these options classes.
	"""

	def update_using_dict(self, update_dict : dict):
		"""Update the current object with the passed dictionary

		Args:
			update_dict (dict): The dictionary to update with
		"""
		#TODO: only set existing keys?
		for key, value in update_dict.items():
			setattr(self, key, value)


	def copy_from(self, other : 'BaseOptions'):
		"""Copy all values from other to self

		Args:
			other (BaseOptions): The other object to copy from
		"""
		for key, value in other.__dict__.items():
			setattr(self, key, value)

	def get_public_attrs_as_str(self, sep = "\n") -> str:
		"""Get all public attributes of the class as a string
		"""
		retval : str = ""
		for key, value in self.__dict__.items():
			if not key.startswith('_'):
				retval += f"{key} : {value}{sep}"
		return retval


	def __getitem__(self, key : str):
		"""Interface for []-itemgetters used in MVTS datasets, if key (class property) not found, returns None

		Args:
			key (str): The key of the item

		Returns:
			(typing.Any) : None if key not found, or the item if key found
		"""
		if key in self.__dict__:
			return self.__dict__[key]
		else:
			return None

	def __setitem__(self, key, value):
		"""Set the value of a property.

		Args:
			key (str): Name of the property
			value (any): Value to be set
		"""
		setattr(self, key, value)



"""
Implements a RunQueue model that synchronizes with text-output from running items in the runqueue.
"""


import datetime
import logging
import threading
import time
import typing
from collections import OrderedDict

from PySide6 import QtCore, QtGui, QtWidgets
from pyside6_utils.models.console_widget_models.console_model import BaseConsoleItem
from configurun.classes.run_queue import RunQueue

log = logging.getLogger(__name__)



class MethodExecutionThread(QtCore.QThread):
	"""Very simple thread that runs a function in a separate thread and stores the return value"""
	def __init__(self, function, *args, **kwargs):
		super().__init__()
		self._function = function
		self._args = args
		self._kwargs = kwargs
		self._ret_val = None



	def run(self) -> typing.Any:
		try:
			self._ret_val = self._function(*self._args, **self._kwargs)
		except Exception as exception: #pylint: disable=broad-exception-caught
			print(f"Error when running function {self._function.__name__} in thread: {exception}")
		self.quit()


class RunQueueConsoleItem(BaseConsoleItem):
	"""
	Model that synchronizes with text-output from running items in the runqueue.
	"""
	loadedLinesChanged = QtCore.Signal(list, int) #Emits all lines that have been changed, together with the start
		# line-index
	dataChanged = QtCore.Signal() #When the metadata of the item changes (e.g. last-edit-date, name, running-state)
	filledTextPositionsChanged = QtCore.Signal(object) #Emitted when the receieved text is filled in the buffer

	def __init__(self,
	    	item_id :int,
			name : str,
			path : str | None = None,
			active_state : bool = True,
			max_buffered_lines : int = 10_000, #Max 10k lines buffered from file
			# max_buffer_size : int = 200_000, #How many characters to read (at max)
			# max_buffer_emit_size : int = 200_000 #How many character to emit (at max) when the buffer changes. Should be
			# 	#> the largest amount of characters that can be added in a single commandline-output signal
			# 	#Note that we can temporarily turn this off when calling on_commandline_output (e.g. when resetting)
			):
		super().__init__()
		self._edit_mutex = threading.Lock() #When adding from/to the buffer, lock this mutex
		self._max_buffered_lines = max_buffered_lines
		# self._current_text : str = ""
		self._loaded_lines : list[str] = [] #List of all lines that have been loaded from the file
		self._total_loaded_lines : int = 0 #Total amount of lines that have been loaded from the file

		self._currently_loaded_line_range : list[int] = [0, 0]
		self._name : str = name
		self._last_edited : datetime.datetime | None = None
		self._id : int = item_id
		self._path : str | None = path
		# max_buffer : int = 100_000
		self._current_gap : list[int] = [0, 0] #We allow for 1 gap in the text
		# self._max_buffer_emit_size = max_buffer_emit_size
		# self._max_buffer_size = max_buffer_size

		self._loaded_buffer : list[int] = [0, 0]

		# self._buffer_startpos : int = 0 #The start position of the buffer (e.g. the first char in the buffer relative
		# 	#to the file it is saved in)


		self.runqueue_console_connection = None

		self._active_state = active_state #By default - assume that the item is running

		self._console_icon = QtGui.QIcon()
		self._console_icon.addFile(":/Icons/icons/actions/Console.png",
	       QtCore.QSize(),
		   QtGui.QIcon.Mode.Normal,
		   QtGui.QIcon.State.Off)

		self._done_console_icon = QtGui.QIcon()
		self._done_console_icon.addFile(":/Icons/icons/actions/Console checked.png",
	       QtCore.QSize(),
		   QtGui.QIcon.Mode.Normal,
		   QtGui.QIcon.State.Off)

		self._threadlist_mutex = threading.Lock()
		self._threads = []




	def get_id(self) -> int:
		"""Returns the id of this item - id is set at creation and should be unique for each item"""
		return self._id

	def set_active_state(self, new_active_state : bool):
		"""Sets the active state of the item (e.g. whether it is running or not)

		Args:
			new_active_state (bool): The new state
		"""
		self._active_state = new_active_state
		self.dataChanged.emit()

	def get_active_state(self) -> bool:
		"""Returns the active state of the item (e.g. whether it is running or not)

		Returns:
			bool: The active state of the item
		"""
		return self._active_state


	def on_commandline_output(self,
			   					item_id : int,
								name : str,
								output_path : str, #pylint: disable=unused-argument
								edit_dt : datetime.datetime,
								# from_line : int, #The line-index of the first line (in the actual file)
								lines : list[str],
								emit_datachanged : bool = True
							):
		"""
		TODO: we assume the lines are ALWAYS passed in-order and append eachother - maybe allow for out of order
			passing by also providing a from-line index?
		Args:
			item_id (int): The id of the item for which the output is received
			name (str): The name of the item
			output_path (str): The path of the item
			edit_dt (datetime.datetime): The last (known) change to the output file
			lines (list[str]): The new text lines
			emit_datachanged (bool, optional): Whether to emit the dataChanged signal if the metadata changed.
				Defaults to True.
		"""
		#=============== First update the log-metadata==================
		if item_id != self._id: #Skip if not the correct id
			return
		if name != self._name:
			raise ValueError(f"Name mismatch: {name} != {self._name} but id is the same - should not happen")

		data_changed = False #Whether the node-data itself changed (e.g. name (namechange not implemented) or edit_dt

		with self._edit_mutex:
			if self._last_edited is None or self._last_edited < edit_dt:
				self._last_edited = edit_dt
				data_changed = True

		if len(lines) > self._max_buffered_lines: #Don't just append useless new lines that will be removed anyway
			lines = lines[-self._max_buffered_lines:]

		self._loaded_lines.extend(lines) #Add the new lines to the loaded lines
		self._total_loaded_lines += len(lines)

		if len(self._loaded_lines) > self._max_buffered_lines: #If we have too many lines, remove the oldest ones
			self._loaded_lines = self._loaded_lines[-self._max_buffered_lines:]

		if data_changed and emit_datachanged: #If the metadata changed, emit the dataChanged signal
			self.dataChanged.emit()

		self.loadedLinesChanged.emit(lines, self._total_loaded_lines-len(lines)) #Emit the new lines



	def get_current_line_list(self) -> typing.Tuple[list[str], int]:
		return self._loaded_lines, 0


	def data(self, role : QtCore.Qt.ItemDataRole, column : int = 0):
		if role == QtCore.Qt.ItemDataRole.DisplayRole:
			if column == 0 :
				return f"{self._id}: {self._name}"
			elif column == 1 :
				return self._last_edited.strftime("%Y-%m-%d %H:%M:%S") if self._last_edited is not None else ""
			elif column == 2 :
				return self._path
			else:
				raise ValueError(f"Invalid column for ConsoleStandardItem (max=2 but is {column})")
		elif role == QtCore.Qt.ItemDataRole.EditRole:
			if column == 0 :
				return f"{self._id}: {self._name}"
			elif column == 1 :
				return self._last_edited
			elif column == 2 :
				return self._path
			else:
				raise ValueError(f"Invalid column for ConsoleStandardItem (max=2 but is {column})")
		# elif role == QtCore.Qt.ItemDataRole.DecorationRole:
		# 	return self._id
		elif role == QtCore.Qt.ItemDataRole.DecorationRole and column == 0: #if first column -> return active/inactive icon
			return self._console_icon if self._active_state else self._done_console_icon
		else:
			return None


class RunQueueConsoleModel(QtCore.QAbstractItemModel):
	"""
	A RunQueue model that synchronizes with text-output from running items in the runqueue by requesting all text-data
	from the runqueue upon connection, and then subscribing to any changes to the text-output of the items
	running in the runqueue.

	On reset, will try to fetch the history of the consoles using the runqueue fetch methods.
	"""


	def __init__(self,
	    		parent: QtWidgets.QWidget | None = None,
				max_initial_console_history : int = 200_000
			) -> None:
		"""Initializes the runqueueconsolemodel

		Args:
			parent (QtWidgets.QWidget | None, optional): The parent. Defaults to None.
			max_initial_console_history (int, optional): Limits the max amount of characters that the model will
				fetch on-reset (per item) this makes sure that the initial load times on model-reset do not become
				too outrageous by only loading part of the data. Especially useful useful when working over network
				connections. NOTE: this limit is not enforced when gradually loading data from the runqueue as
				we assume that it's not feasible to load too much text-data in a single session.
				Defaults to 200_000. #TODO: probably better to do this anyway as creating new strings for each
				append can be very expensive if the string is very large.
				-1 for no limit.
		"""

		super().__init__(parent)
		self._run_queue = None
		self._id_item_map_mutex = threading.Lock()
		self._id_item_map : typing.OrderedDict[int, RunQueueConsoleItem] = OrderedDict({}) #Maps run queue id to item
			# Note: we use an ordered dict to keep a consistent item-order for the UI
		self._ignored_ids : typing.Set[int] = set() #Ids that are ignored/not tracked

		self._max_initial_console_history = max_initial_console_history

		self._new_cmd_text_signal : QtCore.SignalInstance | None= None
		self._active_ids_signal : QtCore.SignalInstance | None = None
		self._queue_reset_signal : QtCore.SignalInstance | None = None

	def set_run_queue(self, run_queue : RunQueue):
		"""Set the runqueue to synchronize with

		Args:
			run_queue (RunQueue): The new runqueue to synchronize with
		"""
		self._run_queue = run_queue
		if self._new_cmd_text_signal is not None:
			self._new_cmd_text_signal.disconnect()
		if self._active_ids_signal is not None:
			self._active_ids_signal.disconnect()
		if self._queue_reset_signal is not None:
			self._queue_reset_signal.disconnect()

		self._active_ids_signal = self._run_queue.currentlyRunningIdsChanged.connect(self.running_ids_changed)
		self._new_cmd_text_signal = self._run_queue.newCommandLineOutput.connect(self.new_command_line_output)
		self._queue_reset_signal = self._run_queue.resetTriggered.connect(self.reset) #Upon queue reset ->
			# re-request all data from the runqueue

	def reset(self, reset_ignored_ids : bool = False):
		"""
		Re-request all data from runQueue and fill model
		#TODO: maybe create a soft-reset that only checks if there is any missing data e.g. when reconnecting to
		a disconnected runqueue - not all data is neccesarily invalid
		"""
		log.info("Resetting RunQueueConsoleModel - re-requesting all console-output from runQueue")
		self.beginResetModel()
		with self._id_item_map_mutex:
			self._id_item_map = OrderedDict({})
		if reset_ignored_ids:
			self._ignored_ids = set()

		if self._run_queue is not None:
			cur_queue = self._run_queue.get_command_line_info_list()
			if cur_queue:
				for cur_id, (name, path, file_size, running) in cur_queue.items():
					if cur_id in self._ignored_ids:
						continue
					item = RunQueueConsoleItem(cur_id, name, path, running)
					self._append_row(item)

					#if limit to initial console history:
					if self._max_initial_console_history > 0 and file_size > self._max_initial_console_history: 
						all_txt, last_edit_dt = self._run_queue.get_command_line_output(
							cur_id, -1, self._max_initial_console_history) #TODO: this probably goes wrong if we resume
							# logging on an existing logfile, as we will "start again" at filepos 0
							# while the actual filepos = file_size-self._max_initial_console_history.
							# probably best to put a start_pos inside run_queue_item
							#TODO: just
					else: #Fetch all data NOTE: this can be very costly, especially over network connections
						all_txt, last_edit_dt = self._run_queue.get_command_line_output(cur_id, -1, file_size)

					lines = all_txt.splitlines()
					item.on_commandline_output( #Append all text to the item #TODO: do this before reading file to avoid
							# missing any of the data that was added during reading
						item_id=cur_id,
						name=name,
						output_path=path,
						edit_dt=last_edit_dt,
						lines=lines,
						emit_datachanged=False #Don't emit changes, otherwise we will try to update items that are not
							# yet known to the model
					) #TODO: maybe only import active items?



		self.endResetModel()
		log.info("Done resetting RunQueueConsoleModel")


	def get_ignored_ids(self) -> typing.List[int]:
		"""
		Get the ids that are ignored/not tracked.
		Returns:
			list[int]: A list of ids that are ignored/not tracked
		"""
		return list(self._ignored_ids)

	def new_command_line_output(self,
			  				item_id : int,
							item_name : str,
							item_path : str,
							item_edit_dt : datetime.datetime,
							item_filepos : int, #pylint: disable=unused-argument
							item_msg : str):
		"""added when new command line output is received

		Args:
			item_id (int): The id of the item for which the output is received
			item_name (str): The name of the item
			item_path (str): The path of the item
			item_edit_dt (datetime.datetime): The last (known) change to the output file
			item_filepos (int): The start-position where the new text should be inserted
			item_msg (str): The new text message
		"""
		if item_id in self._ignored_ids:
			return
		if item_id not in self._id_item_map:
			item = RunQueueConsoleItem(item_id, item_name, item_path)
			self.append_row(item)

		#TODO: maybe process in non-main thread? Settext probably most intensive though...
		self._id_item_map[item_id].on_commandline_output(
			item_id,
			item_name,
			item_path,
			item_edit_dt,
			lines=item_msg.splitlines(),
			# item_filepos,
		)

	def running_ids_changed(self, runnings_ids : typing.List[int]):
		"""Called when the running ids changed, updates the console-items to reflect the active/inactive state"""
		with self._id_item_map_mutex:
			for cur_id in self._id_item_map:
				self._id_item_map[cur_id].set_active_state(cur_id in runnings_ids)


	def columnCount(self, parent : QtCore.QModelIndex = QtCore.QModelIndex()) -> int: #pylint: disable=unused-argument
		return 3

	def rowCount(self, parent : QtCore.QModelIndex = QtCore.QModelIndex()) -> int: # pylint: disable=unused-argument
		return len(self._id_item_map)

	def un_ignore_id(self, item_id : int):
		"""Un-ignore the item with the passed id. Re-synchronizes text-data with the runqueue and re-adds the item
		to the model for the view to display
		TODO: We request the whole text file, but this might take a long while for very large files. Instead
		we should only request the missing data (e.g. the last 1000 chars) and append that to the current text
		Args:
			item_id (int): The item to un-ignore
		"""
		log.debug(f"Un-ignoring id {item_id}")
		with self._id_item_map_mutex:
			if item_id not in self._ignored_ids:
				return #Already un-ignored
			self._ignored_ids.remove(item_id)

		if self._run_queue is not None:
			cur_queue = self._run_queue.get_command_line_info_list()
			if item_id in cur_queue:
				name, path, file_size, currently_running = cur_queue[item_id]
				item = RunQueueConsoleItem(item_id, name, path, currently_running)
				self.append_row(item)
				self.dataChanged.emit(self.index(0,0), self.index(self.rowCount()-1, self.columnCount()-1))

				#TODO: instead of re-requesting all data, only request the missing data - don't delete upon ignore,
				#instead, simply don't show/update the item in the model
				all_txt, last_edit_dt = self._run_queue.get_command_line_output(item_id, -1, file_size)
				lines = all_txt.splitlines()
				item.on_commandline_output( #Append all text to the item
					item_id=item_id,
					name=name,
					output_path=path,
					edit_dt=last_edit_dt,
					lines=lines
				) #TODO: maybe only import active items?

		# item = RunQueueConsoleItem()

	# def reset(self):
	# 	"""Re-request all data and refill model"""
	# 	self.beginResetModel()
	# 	self._id_item_map = {}
	# 	if self._run_queue is not None:
	# 		cur_queue = self._run_queue.get_command_line_info_list()

	# 		for id, (name, path, file_size) in cur_queue.items():
	# 			if id in self._ignored_ids:
	# 				continue
	# 			item = RunQueueConsoleItem(id, name, path)
	# 			with self._id_item_map_mutex:
	# 				self._id_item_map[id] = item
	# 			self.appendRow(item)
	# 	self.endResetModel()

	def _append_row(self, item : RunQueueConsoleItem):
		"""Append a row to the model - consisting of a single ConsoleStandardItem.
		NOTE: this is the same as appendRow, but without the begin/endInsertRows calls, better to be used during model
		reset to avoid emitting dataChanged/rowInserted signals for new items as this might cause issues when
		begin/endresetModel is called during a reset

		Args:
			item (ConsoleStandardItem): The item to append (each row corresponds to 1 item)
		"""
		cur_item_id = item.get_id()
		with self._id_item_map_mutex:
			if cur_item_id in self._id_item_map:
				log.warning("Item with this ID already in model, duplicate ID? Skipping row-insertion.")
				return
			self._id_item_map[cur_item_id] = item
			#TODO: only emit dataChanged for this item
		item.dataChanged.connect(
			# lambda *args: print(f"Data changed for item {item._id}")
			lambda item_id=cur_item_id: self.dataChanged.emit(
				# self.index(0, 0),
				# self.index(self.rowCount()-1, self.columnCount()-1))
				self.index(list(self._id_item_map.keys()).index(item_id), 0),
				self.index(list(self._id_item_map.keys()).index(item_id), self.columnCount()-1))
		)



	def append_row(self, item : RunQueueConsoleItem):
		"""Append a row to the model - consisting of a single ConsoleStandardItem
		Args:
			item (ConsoleStandardItem): The item to append (each row corresponds to 1 item)
		"""
		self.beginInsertRows(QtCore.QModelIndex(), self.rowCount(), self.rowCount()) #No parent, insert at end
		self._append_row(item)
		self.endInsertRows()

	def removeRow(self, row: int, parent : QtCore.QModelIndex) -> None:
		self.beginRemoveRows(parent, row, row)
		item_id = list(self._id_item_map.keys())[row]
		with self._id_item_map_mutex:
			del self._id_item_map[item_id]
			self._ignored_ids.add(item_id)

		ignored_list = ", ".join([str(i) for i in list(self._ignored_ids)])
		log.info(f"Removed row with id {item_id} - ignorelist: {ignored_list}")
		self.modelReset.emit() #Why is this needed?

	def parent(self, index : QtCore.QModelIndex) -> QtCore.QModelIndex: #pylint: disable=unused-argument
		return QtCore.QModelIndex() #No parents

	def data(self, index : QtCore.QModelIndex, role : QtCore.Qt.ItemDataRole = QtCore.Qt.ItemDataRole.DisplayRole):
		#Check if index is valid
		if not index.isValid(): #if index is not valid, return None
			return None

		#Get the item from the index
		item = index.internalPointer()
		assert(isinstance(item, RunQueueConsoleItem)), "Invalid item type in RunQueueConsoleModel"

		if role == QtCore.Qt.ItemDataRole.DisplayRole or role == QtCore.Qt.ItemDataRole.EditRole:
			# return item.data(index.column()) #Return the data (str) of the item
			return item.data(role=role, column=index.column())
		elif role == QtCore.Qt.ItemDataRole.ToolTipRole: #When hovering - also display the full name
			return item.data(role=QtCore.Qt.ItemDataRole.DisplayRole, column=index.column())
		elif role == QtCore.Qt.ItemDataRole.DecorationRole:
			return item.data(role=role, column=index.column())
		elif role == QtCore.Qt.ItemDataRole.UserRole + 1:
			return item
		else:
			return None
		# return super().data(index, role)

	def index(self, row : int, column : int, parent : QtCore.QModelIndex = QtCore.QModelIndex()) -> QtCore.QModelIndex:
		"""Return the index of the item in the model specified by the given row, column and parent index.

		Args:
			row (int): The row of the item
			column (int): The column of the item
			parent (QtCore.QModelIndex, optional): The parent index. Defaults to QtCore.QModelIndex().

		Returns:
			QtCore.QModelIndex: The index of the item
		"""
		if not parent.isValid(): #If top-level item (should be all items actually)
			id_keys = list(self._id_item_map.keys())
			return self.createIndex(row, column, self._id_item_map[id_keys[row]])
		else: #If item -> no children
			return QtCore.QModelIndex()


if __name__ == "__main__":
	#pylint: disable=ungrouped-imports, protected-access, redefined-outer-name, missing-function-docstring, invalid-name
	from pyside6_utils.widgets.console_widget import ConsoleWidget
	app = QtWidgets.QApplication([])
	model = RunQueueConsoleModel()
	print("Now running a test-instance of RunQueueConsoleModel")

	widget = ConsoleWidget(ui_text_min_update_interval=0.05)
	# widget = QtWidgets.QTableView()
	widget.set_model(model)
	widget.show()
	model.append_row(RunQueueConsoleItem(1, "test1"))
	# app.exec()
	model.new_command_line_output(
		item_id=2,
		item_name="test2",
		item_path="",
		item_edit_dt=datetime.datetime.now(),
		item_filepos=0,
		item_msg = "testmsg2"
	)
	samedt = datetime.datetime.now()
	model.new_command_line_output(
		item_id=3,
		item_name="test3",
		item_path="",
		item_edit_dt=samedt,
		item_filepos=0,
		item_msg = "testmsg3"
	)
	model.new_command_line_output(
		item_id=4,
		item_name="test4",
		item_path="",
		item_edit_dt=samedt,
		item_filepos=0,
		item_msg = "testmsg4"
	)
	model._id_item_map[4]._active_state = False
	cur_text = "\n".join([f"This is a test {i}" for i in range(1, 1_000_000)]) + '\nThis should now continue...\n'


	def testfunc(model : RunQueueConsoleModel, start_pos):
		i = 10
		cur_pos = start_pos
		while True:
			i+= 1
			new_msg = f"Currently logging {i}\n\n"
			model.new_command_line_output(
				item_id=1,
				item_name="test1",
				item_path="",
				item_edit_dt=datetime.datetime.now(),
				item_filepos=cur_pos,
				item_msg = new_msg
			)
			cur_pos += len(new_msg)
			time.sleep(0.01)

	# print("len curtext: ", len(cur_text))
	# model.new_command_line_output(
	# 	item_id=1,
	# 	item_name="test1",
	# 	item_path="",
	# 	item_edit_dt=datetime.datetime.now(),
	# 	item_filepos=0,
	# 	item_msg = cur_text
	# )
	# model.new_command_line_output(
	# 	item_id=1,
	# 	item_name="test1",
	# 	item_path="",
	# 	item_edit_dt=datetime.datetime.now(),
	# 	item_filepos=len(cur_text),
	# 	item_msg = "This should be right behind the cur_text string"
	# )
	test_thread = threading.Thread(target=lambda: testfunc(model, len(cur_text)))
	test_thread.start()

	def func(model):
		time.sleep(4)
		model.new_command_line_output(
			item_id=1,
			item_name="test1",
			item_path="",
			item_edit_dt=datetime.datetime.now(),
			item_filepos=0,
			item_msg = cur_text
		)
	thread2 = threading.Thread(target=lambda: func(model))
	thread2.start()
	print("Now executing app")
	app.exec()



"""
Implements several option-dataclasses to show how to use this framework.
"""
import typing
from dataclasses import dataclass, field

from pyside6_utils.classes.constraints import (Interval, StrOptions, ConstrainedList)

from configurun.configuration.base_options import BaseOptions


@dataclass
class ExampleDatasetOptions(BaseOptions):
	"""Example dataset options
	"""

	dataset_name : str | None = field(default=None, metadata=dict(
			display_name="Dataset Name",
			help="Name of the dataset to be used, is used for logging purposes and keeping track of the results",
			constraints = [str, None]
		)
	)


	indexes : typing.List[int] | None = field(
		default=None,
		metadata=dict(
			display_name="Indexes",
			help="Indexes of the dataset to be used (list of ints > 0)",
			constraints = [ConstrainedList([Interval(int, 0, None, closed='both')]), None]
		)
	)


	test_ratio : float = field(
		default=0,
		metadata=dict(
			display_name="Test Ratio",
			help=  		"Set aside this ratio of the dataset as a test set",
			constraints = [Interval(float, 0,1, closed='both')]
		)
	)
	val_ratio : float = field(
		default=0.2,
		metadata=dict(
			display_name="Validation Ratio",
			help= 		"Ratio of the dataset to be used as a validation set",
			constraints = [Interval(float, 0,1, closed='both')]
		)
	)

	data_dir : str = field(
		default="./data",
		metadata=dict(
			display_name="Data Directory", #TODO: make this a path
			help="Data directory path",
			constraints = [str]
		)
	)
	pattern : str | None = field(
		default=None,
		metadata=dict(
			display_name="Pattern",
			help="Regex pattern used to load the files",
			constraints = [str, None]
		)
	)
	val_pattern : str | None = field(
		default=None,
		metadata=dict(
			display_name="Validation File Pattern",
			help=("Regex pattern used to select files contained in `data_dir` for the validation set"),
			constraints = [str, None]
		)
	)

	cross_validation_folds : int | None = field(
		default=None,
		metadata=dict(
			display_name="Cross Validation Folds",
			help=("If specified (>1), will perform cross-validation with the specified number of folds."),
			constraints = [Interval(type=int, left=1, right=None, closed="both"), None]
		)
	)

	test_pattern : str | None = field(default=None, metadata=dict(
			display_name="Test File Pattern",
			help=("Regex pattern used to select files inside data_dir for the test set"),
			constraints = [str, None]
	)
	)



@dataclass
class ExtendedExampleDatasetOptions(ExampleDatasetOptions):
	"""Extends the example dataset options with some more options"""
	labels : typing.List[str] | None = field(
		default=None,
		metadata=dict(
			display_name="Labels",
			help=("In case a dataset contains several labels (multi-task), "
				   "which type of labels should be used in regression or classification, the name of column(s)."),
			constraints = [ConstrainedList([str]), None] #TODO: list of strings?
	))
	normalization : typing.Literal['standardization', 'minmax', 'per_sample_std', 'per_sample_minmax'] = field(
		default='standardization',
		metadata=dict(
			display_name="Normalization",
			help="If specified, will apply normalization on the dataset.",
			display_path="Normalization",
			constraints = [StrOptions({'standardization', 'minmax', 'per_sample_std', 'per_sample_minmax'}), None]
		)
	)

	norm_from : str | None = field(
		default=None,
		metadata=dict(
			display_name="Load Settings from file",
			help=("If given, will read normalization values from the specified file instead of computing them."),
			display_path="Normalization",
			constraints = [str, None]
		)
	)


@dataclass
class ExampleGeneralOptions(BaseOptions):
	"""
	Example of general options - not specific to any model or dataset
	"""

	## Run from command-line arguments
	# I/O
	output_dir : str = field(
		default='output',
		metadata=dict(
			display_name="Output Dir",
			help="Output directory",
			display_path="Output Settings"
		)
	)

	experiment_name : str = field(
		default='',
		metadata=dict(
			display_name="Experiment Name",
			help=("A string identifier/name for the experiment to be run - it will be appended to the output directory"
				"name, before the timestamp"),
			display_path="Output Settings"
		)
	)

	comment : str = field(
		default='',
		metadata=dict(
			display_name="Comment",
			help="A comment/description for this experiment."
		)
	)

	print_interval : int = field(
		default=1,
		metadata=dict(
			display_name="Print Interval",
			help="Print every x interval"
		)
	)

	gpu_indexes : typing.List[int] | None = field(
		default=None,
		metadata=dict(
			display_name="GPU indexes",
			help="Indexes of the videocards to use - None for CPU, -1 for all",
			constraints = [ConstrainedList([Interval(int, -1, None, closed='both')]), None]
		)
	)

	processes : int = field(
		default=-1,
		metadata=dict(
			display_name="Number of processes",
			help="Number of processes to use."
		)
	)
	threads : int = field(
		default=0,
		metadata=dict(
			display_name="Threads",
			help="Number of threads to use"
		)
	)
	seed : int | None = field(
		default=None,
		metadata=dict(
			display_name="Seed",
			help="Seed to use for the process"
		)
	)


@dataclass
class ExampleModelOptions(BaseOptions):
	"""Simple example model settings"""

	simple_int_variable : int = 5

	int_model_param : int = field(
		default=1,
		metadata=dict(
			display_name="Integer Model Parameter",
			help="An integer parameter for the model",
			constraints = [int]
		)
	)

	string_model_param : str = field(
		default="",
		metadata=dict(
			display_name="String Model Parameter",
			help="A string parameter for the model",
			constraints = [str]
		)
	)
	float_in_range_model_param : float = field(
		default=0.5,
		metadata=dict(
			display_name="Float In Range Model Parameter",
			help="A float parameter for the model between 0 and 1",
			constraints = [Interval(float, 0,1, closed='both')]
		)
	)

@dataclass
class ExtendedExampleModelOptions(ExampleModelOptions):
	"""Adds some more options to the example model"""
	extended_str_param : str = field(
		default="",
		metadata=dict(
			display_name="Extra Str Parameter",
			help="An extra string parameter for the model",
			constraints = [str]
		)
	)

	extended_int_param : int = field(
		default=1,
		metadata=dict(
			display_name="Extra Int Parameter",
			help="An extra integer parameter for the model",
			constraints = [int]
		)
	)

	extended_choice_param : typing.Literal["choice1", "choice2"] = field(
		default="choice1",
		metadata=dict(
			display_name="Extra Choice Parameter",
			help="An extra choice parameter for the model",
			contraints_help = {"choice1": "Description of Choice 1",
		      "choice2": "Description of Choice 2"},
		)
	)



@dataclass
class ExampleMainOptions(BaseOptions):
	"""
	Contains the main options that determine the rest of the option-classes. The arguments in this class are parsed
	first, and then the other classes are chosen and parsed based on what is found in this class
	"""
	model_type : typing.Literal["ExampleModel", "ExtendedExampleModel"] | None = field(
		default=None,
		metadata=dict(
					#TODO: add a required flag
			display_name="Model Type",
			constraints_help= {
				"ExampleModel": "A simple example model",
				"ExtendedExampleModel": "An extended example model, extends ExampleModel"
			},
			help="Model(class) to use",
			required=True
		)
	)

	dataset_type : typing.Literal['ExampleDataset', "ExtendedExampleDataset"] | None = field(
		default=None,
		metadata=dict(
			display_name="Dataset Type",
			help="""Which type of dataset is going to be used.""",
			constraints_help= {
				"ExampleDataclass": "A simple example dataset",
				"ExtendedExampleDataclass": "An extended example dataset, extends exampleDataclass"
			},
			required=True
		)
	)



"""
Implements ConfigurationData dataclass - this is a dataclass that holds the data for the actual Configuration class.
The configuration-class is a wrapper around the ConfigurationData class that provides convenience functions for
loading/saving options from/to a file and for creating a new options class from a set of passed options as well
as QT signals/slots for updating the GUI when the options are changed. 

"""

from dataclasses import dataclass, field
import typing
from configurun.configuration.base_options import BaseOptions


@dataclass
class Configuration(object):
	"""
	A Configuration object is built up using several (sub-configuration / option) dataclass instances.
	This class is a wrapper around a dict of these sub-options classes. The __getattr__ function is used to
	allow the user to access any of the sub-options classes using the . operator.

	The individual option-instances are stored in the 'options' dict. The keys of this dict are the names of the
	sub-options classes, e.g. "model_options", "dataset_options", "training_options", etc.


	NOTE: due to the way the __getattr__ function is implemented, the sub-options classes must not have overlapping
	attribute names.
	"""
	#Use factory to create empty dict
	options : typing.Dict[str, BaseOptions | None] = field(default_factory=dict)
	 #Dict in which the sub-option dataclasses are stored. Example of a
		# A possible configuration of this dict is:
		# {
		# 	"main_options": MainOptions,
		# 	"general_options": GeneralOptions,
		# 	"model_options": ModelOptions,
		# 	"dataset_options": DatasetOptions,
		# 	"training_options": TrainingOptions
		# }
		# Where each of the instances are dataclasses that inherit from BaseOptions

	def get_option_types(self) -> typing.Dict[str, type[BaseOptions] | type[None]]:
		"""Returns a dict of the current option-class types
			E.g.:
			{
				"model_options": SklearnModelOptions,
				"dataset_options": BaseDatasetOptions,
				"training_options": SklearnTrainingOptions
			}
		"""
		return {key: type(value) for key, value in self.options.items()}

	def hasinstance(self, instance_type: type) -> bool:
		"""
		Check if Configuration.options has an instance of the given type
		E.g.
		self.options = {
			"model_options": SklearnModelOptions,
			"dataset_options": BaseDatasetOptions,
		}
		self.hasinstance(SklearnModelOptions) -> True
		self.hasinstance(BaseDatasetOptions) -> True
		self.hasinstance(SklearnTrainingOptions) -> False
		"""
		for value in self.options.values():
			if isinstance(value, instance_type):
				return True
		return False
		# return any(isinstance(value, instance_type) for value in self.options.values())

	def hasattr(self, key):
		"""
		Check if one of the sub-options classes has the given attribute
		"""

		for options_class in self.options.values():
			if hasattr(options_class, key):
				return True
		return False

	def __getattr__(self, key : str):
		"""
		Allows the 'OptionsData' class to be used as a flattened dictionary / by using the . operator.
		This is useful when using this options class in training code as we can typehint to one of
		the option-classes and address the options using one . operator, e.g.:
		config.model_options.model_type becomes config.model_type

		For typehinting, we can use the typing.Union type, e.g.:
		def train_model(config : typing.Union[SklearnTrainingOptions, MvtsTrainingOptions]):
			#Do something with config.model_type


		NOTE: the limitation to this is that the suboptions classes must not have overlapping attribute names
		This is why validateSubOptions should be called when chaning suboptions
		TODO: check for overlapping attribute names in the __init__ function?
		"""
		if key.startswith("__"): #Return special/internal attributes, otherwise things like copy will go wrong
			raise AttributeError(f"Attribute {key} not found in any of the options classes")

		for options_instance in self.options.values():
			if hasattr(options_instance, key):
				return getattr(options_instance, key)

		raise AttributeError(f"Attribute {key} not found in any of the options classes")

	def __getitem__(self, key):
		return self.__getattr__(key)
	
	def get(self, key : str, default : typing.Any):
		"""
		A safe alternative to the __getattr__ function. This function will return the default value if the given
		key is not found.

		Args:
			key (str): The attribute to look for
			default (typing.Any): The default value to return if the attribute is not found
		"""
		if key is None:
			return default
		try:
			return self.__getattr__(key)
		except AttributeError:
			return default


	def get_dict(self):
		"""
		Return the full configuration as a dict.
		"""
		new_dict = {}
		for key, value in self.options.items(): #Convert all sub-options instances to dicts
			new_dict[key] = value.__dict__
		return new_dict

	@staticmethod
	def get_configuration_from_passed_options(option_dict : typing.Dict[str, BaseOptions]) -> 'Configuration':
		"""
		Create a new Configuration instance from the passed options.

		Args:
			option_dict (typing.Dict[str, BaseOptions]): A dict of options to add to the configuration. Options must
				inherit from BaseOptions and must have the @dataclass-decorator to be fully compatible with
				the Configuration-ui app. If directly used for the attribute-lookup function of this class,
				this is not neccesry.

				NOTE: all options must have unique attribute names, otherwise the attribute-lookup function will
				return the first attribute it finds with the given name.

				#TODO: enforce this in the __init__ function of this class?
		"""
		new_config = Configuration()
		new_config.options = option_dict #type: ignore
		return new_config


if __name__ == "__main__":
	#Run some tests on the ConfigurationData class to see whether lookup is working
	test_config = Configuration()

	@dataclass
	class TestDataClass(BaseOptions):
		"""Test dataclass"""
		model_type : str = "sklearn"
		comment : str = "This is a test class"


	test_config.options["model_options"] = TestDataClass()
	print(hasattr(test_config, "model_type"))
	print(hasattr(test_config, "option"))
	print(isinstance(test_config, Configuration))
	print(isinstance(test_config, TestDataClass))
	# print(test_config.options)
	# print(test_config.model_type)
	# copy = deepcopy(test_config)



"""
Describes all the data types used for communication between the client and the server
Examples:
	- Transmission: A transmission containing the transmission type and the transmission data
	- TransmissionType: The type of transmission (e.g. public key, login, ...), how to handle transmission data
	- TransmissionData: The data of a transmission (e.g. the public key, the login password, ...)

	- clientData: The data of a client (e.g. the client socket, the client public key, the client session key, ...)
	- serverData: The data of a server (e.g. the server socket, the server private key, ...)

"""

import enum
import socket
import threading
from abc import abstractmethod
from ctypes import c_uint32
from dataclasses import dataclass
import select

import dill
from Crypto.Cipher import AES
from Crypto.PublicKey import RSA
from Crypto.Random import get_random_bytes

RSA_KEY_SIZE_BITS = 2048 #The size of the key used for encryption/decryption
PASSWORD_HASH_SIZE = 512 #The size of the password hash in bytes
MAX_RECV_SIZE = 65_536 #The maximum size of a TCP-package in bytes which we recv per iteration 2^16
AES_KEY_SIZE = 32 #The size of the used AES key in bytes
AES_EMPTY_KEY = b"\x00"*16 #16 bytes of "0" - not used when not encrypted

class AuthenticationException(Exception):
	"""When an error occurs during the authentication process, or when authentication is invalidated
	"""

class TransmissionType(enum.IntEnum):
	"""
	A transmission type - communication between client and server is done by specifying the transmission type,
	followed by the transmission data. The transmission type specifies what kind of data is being sent and how the
	client should handle this data. Also see the ```Transmission``` class.
	"""
	#=========== (Always) Unencrypted transmissions ===========
	PUB_KEY = 0 #Sending of public key - 2048

	#=========== Sometimes encrypted transmissions ===========
	STATE = enum.auto() #A raw state transmission consisting of this unsigned integer, followed the state unsigned
		#integer, followed by a string of size 2048(bits)

	#=========== Always Encrypted transmissions ===========
	SESSION_KEY = enum.auto() #Sending of session key - 256
	LOGIN = enum.auto() #Sending of password
	PICKLED_OBJECT = enum.auto() #Sending of data

class StateMsgType(enum.Enum):
	"""
	If transmissiontype is state, then this enum specifies the type of state message
	"""
	ERROR = 0 #A general error message has been sent
	LOGIN_ERROR = enum.auto() # a loging error has occured
	LOGIN_ACCEPTED = enum.auto() #The login has been accepted
	GENERAL_MSG = enum.auto() #A general message has been sent


class PickledDataType(enum.Enum):
	"""
	If transmissiontype is pickled_object, then this enum specifies the type of pickled data.
	Determines what is done with a TransmissionType.pickled_object transmission
	"""
	METHOD_CALL = 0 #The pickled data is a function call
	METHOD_RETURN = enum.auto() #The pickled data is a function result
	SIGNAL_EMIT = enum.auto() #The pickled data is a signal emit

	def __eq__(self, __value: 'PickledDataType') -> bool:
		"""Overload the default __eq__ function to also work with pickled/unpickled data, don't compare the enum
		hashes to each other, instead compare the name and value of the enum
		"""
		return (type(self).__qualname__ == type(__value).__qualname__) \
			and (self.name == __value.name and self.value == __value.value)



@dataclass
class TransmissionData():
	"""
	Base class for transmissiondata.
	All transmission data should inherit from this class and implement the abstract methods.
	"""
	@staticmethod
	@abstractmethod
	def from_transmission_bytes(transmission_data : bytes) -> 'PubKeyTransmissionData':
		"""
		Parses the transmission data from the given bytes to the object from which it was called. This method should

		Args:
			transmission_data (bytes): The transmission data

		Returns:
			TransmissionData: The parsed transmission data (instance of the class from which this method was called)
		"""
		raise NotImplementedError("Abstract method from_transmission_data should be implemented by subclass")

	@abstractmethod
	def to_transmission_bytes(self) -> bytes:
		"""Converts the transmission data to bytes to be appended to a Transmission() object to be sent using
		a socket

		Returns:
			bytes: the bytes of the transmission data
		"""
		raise NotImplementedError("Abstract method to_transmission_data should be implemented by subclass")

	@property
	@abstractmethod
	def TRANSMISSION_TYPE(self) -> TransmissionType: #pylint: disable=invalid-name
		"""
		Returns the transmission type of the transmission data
		"""
		raise NotImplementedError("Abstract property transmission_type should be implemented by subclass")


@dataclass
class PickleTransmissionData(TransmissionData):
	"""
	A transmission containing a pickled object. Is used as a general-purpose transmission type
	to allow for the transmission of arbitrary objects.
	NOTE: this transmission type should ALWAYS be encrypted using AES, and should only be allowed
	after authentication as it allows for arbitrary code execution on the server-side.

	NOTE: we're using dill instead of pickle as it allows for more flexibility in what can be pickled

	"""

	unpickled_data : object #The unpickled data

	@staticmethod
	def from_transmission_bytes(transmission_data : bytes) -> 'PickleTransmissionData':
		unpickled_data = dill.loads(transmission_data)
		return PickleTransmissionData(unpickled_data)

	def to_transmission_bytes(self) -> bytes:
		return dill.dumps(self.unpickled_data)

	@property
	def TRANSMISSION_TYPE(self) -> TransmissionType:
		return TransmissionType.PICKLED_OBJECT

@dataclass
class PubKeyTransmissionData(TransmissionData):
	"""
	A transmission that contains the public key of the server
	"""
	pubkey : bytes #The public key of the server

	@property
	def TRANSMISSION_TYPE(self) -> TransmissionType:
		return TransmissionType.PUB_KEY

	@staticmethod
	def from_transmission_bytes(transmission_data : bytes) -> 'PubKeyTransmissionData':
		"""Receives a pubkey transmission from the given socket (assumes that the <transmission_type> has already been
		  received which resulted in PubkeyTransmission)
		Args:
			transmission_data (bytes): The data-part of the transmission from which to create the pubkey
		Returns:
			PubkeyTransmission: A pubkey transmission object
		"""
		#Receive the public key
		# assert(len(transmission_data) == RSA_KEY_SIZE_BITS/8) #Should be exactly the size of the public key
		pubkey = transmission_data
		return PubKeyTransmissionData(pubkey)

	def to_transmission_bytes(self) -> bytes:
		"""
		Returns the transmission data of a pubkey transmission
		NOTE: this can never be encrypted as the client does not have the public key of the server yet
		"""
		data=bytearray()
		data.extend(self.pubkey) #Public key
		return data

@dataclass
class LoginTransmissionData(TransmissionData):
	"""A login transmission - contains a password.
	NOTE: SHOULD ALWAYS BE ENCRYPTED WITH AES - as it contains the password hash
	TODO: also hash the password on the client side to prevent server from knowing the password
	"""
	password : str #The password hash of the user

	@property
	def TRANSMISSION_TYPE(self) -> TransmissionType:
		return TransmissionType.LOGIN

	@staticmethod
	def from_transmission_bytes(transmission_data: bytes) -> 'LoginTransmissionData':
		# assert(len(transmission_data) == PASSWORD_HASH_SIZE)
		return LoginTransmissionData(
			password=transmission_data.decode('utf-8') #Decode bytes to string
		)

	def to_transmission_bytes(self) -> bytes:
		"""Returns the transmission data of a pubkey transmission
		"""
		data=bytearray()
		data.extend(self.password.encode('utf-8')) #Encode string to bytes
		return data

@dataclass
class AESSessionKeyTransmissionData(TransmissionData):
	"""A transmission that contains the session key encrypted with the public key of the server
	"""
	encrypted_session_key : bytes #The encrypted session key

	@property
	def TRANSMISSION_TYPE(self) -> TransmissionType:
		return TransmissionType.SESSION_KEY

	@staticmethod
	def from_transmission_bytes(transmission_data: bytes) -> 'AESSessionKeyTransmissionData':
		assert len(transmission_data) == RSA_KEY_SIZE_BITS/8, "Key should be exactly the size of the public key"
		return AESSessionKeyTransmissionData(
			encrypted_session_key=transmission_data
		)

	def to_transmission_bytes(self) -> bytes:
		"""Returns the transmission data of a pubkey transmission
		"""
		data=bytearray()
		data.extend(self.encrypted_session_key)
		return data

@dataclass
class StateTransmissionData(TransmissionData):
	"""Transmits a state of the server/client interaction - e.g. login accepted, login failed, general messages, etc...
	"""
	# TRANSMISSION_TYPE = TransmissionType.STATE # type: ignore
	state_msg_type : StateMsgType #The current state of the transmission
	state_msg : str #A message describing the current state of the transmission -> should be max 2048 bytes/characters

	@property
	def TRANSMISSION_TYPE(self) -> TransmissionType:
		return TransmissionType.STATE

	@staticmethod
	def from_transmission_bytes(transmission_data : bytes) -> 'StateTransmissionData':
		state_msg_type = StateMsgType(c_uint32.from_buffer_copy(transmission_data[:4]).value)
		state_msg = transmission_data[4:].decode('utf-8')
		return StateTransmissionData(state_msg_type, state_msg)

	def to_transmission_bytes(self) -> bytes:
		"""Converts StateTransmission to a byte array used for sending"""
		data = bytearray()
		data.extend(c_uint32(self.state_msg_type.value)) # type: ignore
		data.extend(self.state_msg.encode('utf-8'))
		return data


@dataclass
class Transmission():
	"""All data is sent through sockets using this class. The transmission contains the transmission size, type and
	data. The transmission type specifies what kind of data is being sent and how the client should	parse it.
	"""
	transmission_size : int #The size of the transmission (excluding this field) - determined how many bytes to receive
	transmission_type : TransmissionType #The type of the transmission
	transmission_data : bytes #The raw data of the transmission - defined by the subclass


	@staticmethod
	def receive(
				recv_socket : socket.socket,
	    		aes_cipher_key : bytes | None  = None,
				timeout_seconds : float | None = None
			) -> 'Transmission': #TODO: timeout
		"""Receives the data of the transmission from the given socket
		Args:
			socket (socket.socket): The socket to receive the transmission from
			aes_cipher_key (bytes): The aes-cipher-key used to decrypt the data - if not provided - the data is assumed
			 to be unencrypted (e.g. when receiving a public key)
			timeout_seconds (float): The timeout in seconds to wait for data to be available on the socket
				if < 0 or None, we wait forever. Raises a TimeoutError if no data is available after the given timeout.
		Returns:
			TransmissionType: The type of the transmission
			bytes: The raw data of the transmission

		Raises:
			TimeoutError: If no data is available on the socket after the given timeout
		"""
		if timeout_seconds is not None and timeout_seconds >= 0:
			#Use select to wait for data to be available on the socket
			#NOTE: this is not really necessary as the socket.recv() function will block until data is available
			#		however, this allows us to specify a timeout
			data_available = select.select([recv_socket], [], [], timeout_seconds)
			if not data_available[0]:
				raise TimeoutError("No data available on socket")


		#Receive the transmission data
		transmission_size = recv_socket.recv(4)
		transmission_size = c_uint32.from_buffer_copy(transmission_size).value #Read in uint32, convert to python type

		aes_nonce = recv_socket.recv(16) #Receive the nonce used for AES encryption

		# data = recv_socket.recv(transmission_size) #NOTE: recv does not guarantee that all data is received
		# We loop and use recv() until we have received all data using MAX_PACKET_SIZE as the max size of a packet
		# Also make sure we don't over-read the data, otherwise the next transmission will go wrong
		data = bytearray()
		while len(data) < transmission_size: 
			data.extend(recv_socket.recv(min(MAX_RECV_SIZE, transmission_size - len(data))))


		assert(len(data) == transmission_size),\
			f"Received transmission size {transmission_size} does not match the received data size {len(data)}"

		if aes_cipher_key is not None:
			aes_cipher = AES.new(aes_cipher_key, AES.MODE_EAX, nonce=aes_nonce)
			data = aes_cipher.decrypt(data)


		new_transmission = Transmission(
			transmission_size = transmission_size,
			transmission_type = TransmissionType(
				c_uint32.from_buffer_copy(data[:4]).value #First 4 bytes are the transmission type
			),
			transmission_data = data[4:]
		)
		return new_transmission

	@staticmethod
	def send(send_socket : socket.socket, transmission_data : TransmissionData, aes_cypher_key : bytes | None  = None):
		"""Sends the data of the transmission to the given socket

		Args:
			socket (socket.socket): The socket to send the transmission to
			transmission_data (TransmissionType): The type of the transmission
			aes_cipher_key (any): The cipher key used for aes encryption - note that some types of transmissions
				should always be encrypted (e.g. login, session key, ...) and some should never be encrypted
				(e.g. public key), some assertions are made.
		"""
		bytes_data = transmission_data.to_transmission_bytes()

		if aes_cypher_key is None:
			#Allow only the following transmission types to be unencrypted:
			assert(isinstance(transmission_data, PubKeyTransmissionData) or
	  				isinstance(transmission_data, StateTransmissionData) or #Should also be encrypted with AES when auth.
					isinstance(transmission_data, AESSessionKeyTransmissionData) #Session token is sha-key-encrypted
				), f"Transmission type {transmission_data.TRANSMISSION_TYPE} should always be encrypted with AES when \
					  transmitting, but no AES key was provided"

		cur_transmission = Transmission(
			transmission_size = len(bytes_data) + 4,
			transmission_type = transmission_data.TRANSMISSION_TYPE,
			transmission_data = bytes_data
		)
		cur_transmission._send(send_socket, aes_cypher_key) # pylint: disable=protected-access
			#_send should only be called from within this class


	def _send(self, send_socket : socket.socket, aes_cypher_key : bytes | None  = None):
		"""Sends the transmission to the given socket
		Args:
			socket (socket.socket): The socket to send the transmission to
			aes_cipher (any): The AES cipher used to encrypt the data - if not provided - the data is assumed to be
				unencrypted (e.g. when sending a public key )
		"""
		packet = bytearray()
		packet.extend(c_uint32(self.transmission_size))  # type: ignore

		data = bytearray()
		data.extend(c_uint32(self.transmission_type.value)) # type: ignore
		data.extend(self.transmission_data)

		nonce =  AES_EMPTY_KEY #16 bytes of "0" - not used when not encrypted
		if aes_cypher_key is not None:
			while nonce == AES_EMPTY_KEY: #Generate a new nonce if it is all 0's (not allowed) ->
					#is VERY unlikely to happen though... but just in case.
				nonce = get_random_bytes(16)

			aes_cipher = AES.new(aes_cypher_key, AES.MODE_EAX, nonce=nonce)
			data = aes_cipher.encrypt(data)

		packet.extend(nonce) #Add the nonce to the packet
		packet.extend(data)

		send_socket.sendall(packet) #Packet: 4bytes transmission_size, 16 bytes nonce, transmission_data



@dataclass
class ClientData():
	"""Dataclass describing a socket connection to a client
	"""
	client_socket : socket.socket
	client_public_key : RSA.RsaKey #TODO: not really necessary after authentication as of now
	client_session_aes_key : bytes #The session key used to encrypt data between the client and the server
	client_listening_thread : threading.Thread
	client_listening_thread_stop_flag : threading.Event #Flag used to stop the client listening thread



"""
Network-equivalent of main_window.py
Contains everything neccesary to run the UI in client-mode, enabling the user to connect to a remotely running
RunQueueServer (see MLQueue.classes.RunQueueServer) and run machine learning tasks on it.
"""

import logging
import typing

from PySide6 import QtCore, QtGui, QtWidgets
from pyside6_utils.utility.catch_show_exception_in_popup_decorator import \
    catch_show_exception_in_popup_decorator

from configurun.classes.run_queue_client import RunQueueClient
from configurun.configuration.configuration_model import ConfigurationModel
from configurun.app.main_window import MainWindow
from configurun.app.widgets.network_login_widget import NetworkLoginWidget

log = logging.getLogger(__name__)

class NetworkMainWindow(MainWindow):
	"""
	The main QT window for this app which provides the user with several tools to edit/manage/run machine learning
	settings.


	This is the Networked-equivalent of main_window.py and implements some extra functionality to connect to a remote
	RunQueueServer (see classes.RunQueueServer) and run machine learning tasks on it.

	Should be provided with:
		- A configuration model - manages the creation of new configurations & the ui
		- A run queue - manages the running of the configurations
		- A window - the main window in which the app should be built
		- workspace_path (str, optional) - the default path to use for the configuration, logfiles etc. If empty, or
			folder does not exist, defaults to ~/Configurun/configurations/
		- settings_in_workspace_path (bool, optional) - Whether to store the settings in the workspace path or in the
			default QSettings location. Defaults to True
	"""
	_authenConnectionStateChanged = QtCore.Signal(bool) #Emitted when the connection state changes (True=authenticated)
		# Copy used to make sure that the signal emitted from the run_queue is emitted in the qt-main thread
		# (since it's a non-qt signal using PySignal). 

	def __init__(self,
				configuration_model : ConfigurationModel,
				run_queue_client : RunQueueClient,
				window : QtWidgets.QMainWindow,
				workspace_path : str = "",
				settings_in_workspace_path : bool = True
			) -> None:
		"""
		Args:
			configuration_model (ConfigurationModel): The configuration model which manages updating the ui and creating
			run_queue (RunQueue): The runqueue which manages running the configurations
			window (QtWidgets.QMainWindow): The window in which the app should be built
			workspace_path (str, optional): The base output-path used for the configurations, logfiles etc.
				If empty, or folder does not exist, defaults to ~/Configurun/configurations/
			settings_in_workspace_path (bool, optional): Whether to store the settings in the workspace path or in the default
				QSettings location. Defaults to True
		"""
		super().__init__(
			configuration_model=configuration_model,
			run_queue=run_queue_client,
			window=window,
			workspace_path=workspace_path,
			settings_in_workspace_path=settings_in_workspace_path
		)

		# assert(type(self._run_queue) == RunQueueClient) #Make sure we're using the right type of queue
		self._run_queue : RunQueueClient = self._run_queue #For Type hinting
		assert(isinstance(self._run_queue, RunQueueClient)) #Make sure we're using the right type of queue

		#=================== Network UI ======================
		self.reconnect_button_1, self._task_queue_overlay_msg = self.get_connection_overlay_ui()
		self.reconnect_button_2, self._console_overlay_msg = self.get_connection_overlay_ui()


		self.ml_overlay_widget = self.ui.runQueueOverlayWidget
		self.console_overlay_widget = self.ui.ConsoleOverlayWidget

		self.ml_overlay_widget.set_overlay_widget(self._task_queue_overlay_msg)
		self.console_overlay_widget.set_overlay_widget(self._console_overlay_msg)

		self.console_overlay_widget.set_overlay_hidden(False)
		self.ml_overlay_widget.set_overlay_hidden(False)

		#================== Network-specific menu ================

		self.menubar = self.ui.menubar
		self.connection_menu = QtWidgets.QMenu(self.menubar)
		self.connection_menu.setTitle("Connection")
		self.menubar.addAction(self.connection_menu.menuAction())
		self.open_connection_action = QtGui.QAction("Connection Settings", self.connection_menu)
		self.open_connection_action.setIcon(QtGui.QIcon(":/Icons/icons/places/network-workgroup.png"))
		self.connection_menu.addAction(self.open_connection_action)


		#=========== Connect/Disconnect window ==============
		self.connection_window = QtWidgets.QMainWindow(self.window)
		self.connection_window.setWindowTitle("Connection")
		self.connection_window.setWindowIcon(self.window.windowIcon())

		self.network_connection_parent = QtWidgets.QWidget()
		self.network_connection_widget = NetworkLoginWidget(self.network_connection_parent, self._settings)
		self.connection_window.setCentralWidget(self.network_connection_parent)
		self.server_connection_state_changed(self._run_queue.is_connected_and_authenticated()) #Set initial state


		#========= Link runqueue signals to UI updates ===========
		#NOTE: authenConnectionChanged is a non-main-thread non-qt-signal
		# as such, we must make sure that updates to the UI are done in the main thread by using a single-shot timer
		self._authenConnectionStateChanged.connect(self.server_connection_state_changed)
		self._run_queue.authenConnectionStateChanged.connect(
			lambda connected : self._authenConnectionStateChanged.emit(connected)
		)



		#=========== Link connection view buttons to connection window ==============
		self.network_connection_widget.connectClicked.connect(self.connect_to_server)
		self.network_connection_widget.disconnectClicked.connect(self.disconnect_from_server)
		self.network_connection_widget.cancelClicked.connect(lambda *_: self.connection_window.close())
		self.network_connection_widget.cancelClicked.connect(lambda *_: self.network_connection_widget.save_histories())

		self.network_connection_widget.connectClicked.connect(
			lambda *_: self.network_connection_widget.save_histories()
		)


		#========= Make sure window is shown when (re)connecting ==========
		self.reconnect_button_1.clicked.connect(lambda *_: self.connection_window.activateWindow())
		self.reconnect_button_2.clicked.connect(lambda *_: self.connection_window.activateWindow())
		self.reconnect_button_1.clicked.connect(lambda *_: self.connection_window.show())
		self.reconnect_button_1.clicked.connect(lambda *_: self.connection_window.setFocus())
		self.reconnect_button_2.clicked.connect(lambda *_: self.connection_window.show())
		self.reconnect_button_2.clicked.connect(lambda *_: self.connection_window.setFocus())
		self.reconnect_button_1.clicked.connect(lambda *_: self.connection_window.raise_())
		self.reconnect_button_2.clicked.connect(lambda *_: self.connection_window.raise_())
		#Also move window to front
		self.open_connection_action.triggered.connect(self.connection_window.setFocus())
		self.open_connection_action.triggered.connect(self.connection_window.show)
		self.open_connection_action.triggered.connect(lambda *_: self.connection_window.raise_())



	def initial_run_queue_load(self) -> None:
		"""Since this is a client, don't load any RunQueue on startup."""

	def check_if_running_ask_stop_items_before_close(self) -> bool:
		"""Overloads the main_window function. Normally, the user shuts down the running items before closing the
		app. For clients this is not neccesary.
		"""
		return True

	@catch_show_exception_in_popup_decorator
	def connect_to_server(self, server_ip : str, server_port : str, server_password : str) -> None:
		"""
		Wrapper around the connect function of the run_queue which displays a message box when encountering an exception
		Args:
			server_ip (str): The ip of the server to connect to
			server_port (str): The port of the server to connect to
			server_password (str): The password to use for authentication
		"""
		self._run_queue.connect_to_server(server_ip=server_ip, server_port=int(server_port), password=server_password)
		print("Connected to server!")

	@catch_show_exception_in_popup_decorator
	def disconnect_from_server(self) -> None:
		"""
		Wrapper around the disconnect function of the run_queue which displays a message box on thrown exceptions.
		"""
		self._run_queue.disconnect_clean_server()

	def server_connection_state_changed(self, connected : bool) -> None:
		"""Update the UI to reflect the connection state, e.g. grey-out the task queue on disconnect as to
		indicate to the user that connection has been lost.

		Args:
			connected (bool): The new connection state (true=Authenticated connection)
		"""
		log.info(f"Connection state changed to {connected}, now updating UI...")
		#=========== hide blocking overlays when connected ==========
		self.console_overlay_widget.set_overlay_hidden(connected)
		self.ml_overlay_widget.set_overlay_hidden(connected)


		#Update the connection window
		if connected:
			cur_ip, cur_port, cur_pw = self._run_queue.get_connection_info()
			self.network_connection_widget.client_connected(cur_ip, str(cur_port), cur_pw)
			self.connection_window.close()

			self.run_queue_table_model.reset_model() #Re-request all data from the server
			self.window.statusBar().showMessage(f"Connected to {cur_ip}:{cur_port}", timeout=0) #Show message until next msg

		else:
			self.network_connection_widget.client_disconnected()
			self.connection_window.activateWindow()
			self.connection_window.show() #Goes wrong here
			self.window.statusBar().showMessage("Not Connected", timeout=0) #Show message until next msg




	def close_event(self, event: QtGui.QCloseEvent) -> None:
		self.network_connection_widget.save_histories() #Also save file-edit history
		self.disconnect_from_server() #Disconnect from server
		return super().close_event(event)


	@staticmethod
	def get_connection_overlay_ui() -> typing.Tuple[QtWidgets.QPushButton, QtWidgets.QWidget]:
		"Create a simple ui with a button to reconnect"
		overlay_widget = QtWidgets.QWidget()
		overlay_widget.setLayout(QtWidgets.QVBoxLayout())
		overlay_widget.layout().setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
		overlay_widget.layout().setContentsMargins(0,0,0,0)
		overlay_widget.layout().setSpacing(0)
		overlay_widget.layout().addWidget(QtWidgets.QLabel("No (authenticated) connection to server..."))
		reconnect_btn = QtWidgets.QPushButton("(Re)connect")
		overlay_widget.layout().addWidget(reconnect_btn)
		return reconnect_btn, overlay_widget


def run_example_network_app(log_level : int = logging.INFO) -> None:
	"""
	Runs an example-client that can connect to a running server. Uses the example_deduce_new_option_classes to
	generate example options for the client
	"""

	#pylint: disable=import-outside-toplevel
	import os
	import tempfile
	from configurun.app.create_run import run_client
	from configurun.examples import example_deduce_new_option_classes
	from configurun.app.main_window import APP_NAME

	tempdir = tempfile.gettempdir()
	workspace_path = os.path.join(tempdir, APP_NAME, "Configurun-Client-App-Example")

	run_client(
		options_source=example_deduce_new_option_classes,
		workspace_path=workspace_path,
		log_level=log_level
	)


if __name__ == "__main__":
	run_example_network_app(log_level=logging.DEBUG)



"""
Implements the RunQueueTreeView class that is used to display a RunQueue in a Qt-Treeview
"""
import logging
import typing

from PySide6 import QtCore, QtWidgets
from pyside6_utils.models import ExtendedSortFilterProxyModel

from configurun.classes.run_queue import (RunQueue, RunQueueItemActions,
                                          RunQueueItemStatus)
from configurun.app.models.run_queue_table_model import RunQueueTableModel

log = logging.getLogger(__name__)

class RunQueueTreeView(QtWidgets.QTreeView):
	"""
	Treeview-equivalent used for run-queue items (MLQueueItems)
	"""

	def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None) -> None:
		super().__init__(parent)

		# self.setModel(model)

		#===========Create menu ============
		self._rc_menu = QtWidgets.QMenu(self)
		self._delete_action = self._rc_menu.addAction("Delete")
		self._cancel_action = self._rc_menu.addAction("Cancel")
		self._move_up_action = self._rc_menu.addAction("Move Up")
		self._move_down_action = self._rc_menu.addAction("Move Down")
		self._move_top_action = self._rc_menu.addAction("Move to Top")
		self._stop_action = self._rc_menu.addAction("Stop")
		self._start_action = self._rc_menu.addAction("Run Item")

		self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
		self.customContextMenuRequested.connect(self.custom_menu_requested)

		self.setSortingEnabled(True)

		self._action_dict = {
			RunQueueItemActions.DELETE: self._delete_action,
			RunQueueItemActions.CANCEL: self._cancel_action,
			RunQueueItemActions.MOVEUP: self._move_up_action,
			RunQueueItemActions.MOVEDOWN: self._move_down_action,
			RunQueueItemActions.MOVETOP: self._move_top_action,
			RunQueueItemActions.STOP: self._stop_action,
			RunQueueItemActions.START: self._start_action,
		}

		# =========== Link actions ===========
		for key in self._action_dict: # pylint: disable=consider-using-dict-items
			if key != RunQueueItemActions.STOP:
				self._action_dict[key].triggered.connect(lambda *_, action=key : self.do_action_on_selection(action))

		#Stop is a special case, ask for confirmation
		self._action_dict[RunQueueItemActions.STOP].triggered.connect(lambda *_: self.confirm_stop_current_index())

		self.proxy_model = ExtendedSortFilterProxyModel(self)
		self.proxy_model.setDynamicSortFilter(True)
		self.proxy_model.set_filter_function("status_filter", self.check_if_current_filter_accepts_row)
		#Make sure the dataChanged signal is emitted when the source model changes

		super().setModel(self.proxy_model)


		#Connect double-click
		self._double_click_connection = self.doubleClicked.connect(self._default_on_double_click)
		self._hide_status_list = []

	def check_if_current_filter_accepts_row(self,
				source_row: int,
				source_parent: QtCore.QModelIndex | QtCore.QPersistentModelIndex,
				source_model: QtCore.QAbstractItemModel
			) -> bool:
		"""Check if the passed row should be visible or not

		Args:
			source_row (int): The row to check
			source_parent (QtCore.QModelIndex): The parent of the row to check

		Returns:
			bool: True if the row should be visible, False otherwise
		"""
		#Status column:
		index = source_model.index(source_row, 0, source_parent)
		# data_roles = [role for role in RunQueueTableModel.CustomDataRoles]
		status = index.data(RunQueueTableModel.CustomDataRoles.StatusRole)[0]
		if status in self._hide_status_list:
			return False
		return True

	def set_item_status_filter(self, hide_status_list : list[RunQueueItemStatus] | None = None) -> None:
		"""Sets the list of statuses that should be filtered (hidden)

		Args:
			hide_status_list (list[RunQueueItemStatus], optional): A list of statuses that should be filtered (hidden).
				Defaults to None.
		"""
		if hide_status_list is None:
			hide_status_list = []
		self._hide_status_list = hide_status_list
		self.proxy_model.invalidateFilter()

	def set_whether_status_filtered(self, status : RunQueueItemStatus,  hide_item : bool) -> None:
		"""Sets whether the passed status should be filtered (hidden) or not

		Args:
			status (RunQueueItemStatus): The status to hide or show
			hide_item (bool): Whether to hide the passed status or not
		"""
		if hide_item:
			if status not in self._hide_status_list:
				self._hide_status_list.append(status)
				self.proxy_model.invalidateFilter()
		else:
			if status in self._hide_status_list:
				self._hide_status_list.remove(status)
				self.proxy_model.invalidateFilter()


	def get_item_status_filter(self) -> list[RunQueueItemStatus]:
		"""Returns a list of all statuses that are filtered (hidden)

		Returns:
			list[RunQueueItemStatus]: A list of all statuses that are filtered (hidden)
		"""
		return self._hide_status_list


	def set_double_click_callback(self,
				function : typing.Callable[[QtCore.QModelIndex], None]
			) -> None:
		"""Used to overwrite the current callback when double-clicking on an item.
		E.g. if we first want to try to load the clicked-item, and on fail, do nothing

		NOTE: disconnects any existing connections

		Args:
			function (typing.Callable[[QtCore.QModelIndex], None]): The function to call when double-clicking on an item
				should take a single argument, the index of the double-clicked item
		"""
		if self._double_click_connection is not None:
			self.doubleClicked.disconnect(self._double_click_connection)
		self._double_click_connection = self.doubleClicked.connect(function)

	def _default_on_double_click(self, index: QtCore.QModelIndex) -> None:
		""" Default action when double-clicking on an item in the treeview.
		Set the hightlight-item when double-clicking on an item

		Args:
			index (QtCore.QModelIndex): The double-clicked item
		"""
		cur_id = index.data(RunQueueTableModel.CustomDataRoles.IDRole)
		log.debug(f"Double clicked on index {index.row()} with id {cur_id}")
		cur_model = self.proxy_model.sourceModel()
		if isinstance(cur_model, RunQueueTableModel):
			cur_model.set_highligh_by_id(cur_id)

	# def model(self):
	# 	"""Return the model for this view
	# 	"""
	# 	return self.proxy_model


	def setModel(self, new_model: RunQueueTableModel) -> None:
		"""Set the model for this view

		NOTE: this treeview uses a proxy model, so the passed model is set as the source model of the proxy model
			<this>.getModel() will return the proxy model, not this passed model
		"""
		ret = self.proxy_model.setSourceModel(new_model)
		#Make sure the dataChanged signal is emitted when the source model changes:
		new_model.dataChanged.connect(self.proxy_model.dataChanged) #NOTE: if we don't this, the proxy model won't emit
			# dataChanged signals. This seems to be a bug in Qt, though I could not find a mention of it on 25-06-2023

		return ret
	def confirm_stop_index(self, index : QtCore.QModelIndex) -> None:
		""" Ask user for confirmation to stop the item at the passed index
		Args:
			index (QtCore.QModelIndex): The index to confirm the stop action on
		"""
		msg = QtWidgets.QMessageBox()
		msg.setIcon(QtWidgets.QMessageBox.Icon.Warning)
		msg.setText("Are you sure you want to stop this item?")
		msg.setInformativeText("This will force stop the item process - any unsaved progress of this run will be lost.")
		msg.setWindowTitle("Confirm stop")
		msg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No)
		msg.setDefaultButton(QtWidgets.QMessageBox.StandardButton.No)
		ret = msg.exec()
		if ret == QtWidgets.QMessageBox.StandardButton.Yes:
			return self.do_action_on_index(RunQueueItemActions.STOP, index)


	def confirm_stop_current_index(self) -> None:
		""" Ask user for confirmation to stop the currently selected item
		"""
		index = self.currentIndex()
		return self.confirm_stop_index(index)



	def custom_menu_requested(self, pos : QtCore.QPoint) -> None:
		"""Create a context menu for the possible actions on the item under the mouse

		Args:
			pos (QtCore.QPoint): The position of the mouse
		"""

		#Get the index of the item under the mouse
		index = self.indexAt(pos)
		#If the index is valid, create a menu
		if not index.isValid():
			return

		possible_actions = index.data(RunQueueTableModel.CustomDataRoles.ActionRole)
		for action in self._action_dict: # pylint: disable=consider-using-dict-items
			if action in possible_actions:
				self._action_dict[action].setVisible(True)
			else:
				self._action_dict[action].setVisible(False)
		self._rc_menu.popup(self.mapToGlobal(pos))

	def do_action_on_index(self, action: RunQueueItemActions, index : QtCore.QModelIndex) -> None:
		"""Do the passed action on the passed index

		NOTE: no safety checks are performed, calling stop on an item will force stop it if it's running
		without asking for user confirmation

		Args:
			action (RunQueueItemActions): The action to perform
			index (QtCore.QModelIndex): The index to perform the action on
		"""
		log.debug(f"Trying to perform action {str(action)} on index {index.row()}")
		# cur_model = self.model()
		if not index.isValid():
			log.info(f"Could not perform action {str(action)} on index {index.row()} - index is invalid")
			return
		try:
			self.model().setData(index, action, RunQueueTableModel.CustomDataRoles.ActionRole)
		except Exception as exception: # pylint: disable=broad-exception-caught
			log.error(f"Failed to perform action {str(action)}: {exception}")
			#Create qt message box with this notification
			msg = QtWidgets.QMessageBox()
			msg.setIcon(QtWidgets.QMessageBox.Icon.Critical)
			msg.setText(f"Failed to perform action {str(action)} on selected index ({index.row()})")
			msg.setInformativeText(f"{type(exception).__name__}: {exception}")
			msg.setWindowTitle("Error")
			msg.exec_()

	def do_action_on_selection(self, action: RunQueueItemActions) -> None:
		"""Do the passed action on the currently selected item (singular)

		Args:
			action (RunQueueItemActions): The action to perform
		"""
		log.debug(f"Trying to perform action {str(action)} on selection ({self.currentIndex().row()})")
		index = self.currentIndex()
		if not index.isValid():
			return
		self.do_action_on_index(action, index)
		log.debug("Done performing action on selection")







def run_example_app():
	""" Run an example of the run_queue tree view"""
	#pylint: disable=import-outside-toplevel
	import sys

	from configurun.examples.example_target_function import example_target_function
	app = QtWidgets.QApplication([])
	window = QtWidgets.QMainWindow()
	window.resize(800, 600)
	central_widget = QtWidgets.QWidget()
	window.setCentralWidget(central_widget)
	layout = QtWidgets.QVBoxLayout()
	central_widget.setLayout(layout)

	queue_view = RunQueueTreeView()
	run_queue = RunQueue(
		target_function=example_target_function
	)
	queue_model = RunQueueTableModel(run_queue=run_queue)

	queue_view.setModel(queue_model)
	layout.addWidget(queue_view)

	for i in range(10):
		run_queue.add_to_queue(
			f"Testitem {i}",
			config={"test": i}, #nonsense config #type: ignore
		)

	window.show()
	res = app.exec()
	log.info("Done!")
	sys.exit(res)


if __name__ == "__main__":
	logging.getLogger('matplotlib').setLevel(logging.INFO)
	logging.getLogger('PySide6').setLevel(logging.DEBUG)
	formatter = logging.Formatter("[{pathname:>90s}:{lineno:<4}] {levelname:<7s}   {message}", style='{')
	handler = logging.StreamHandler()
	handler.setFormatter(formatter)
	logging.basicConfig(
		handlers=[handler],
		level=logging.DEBUG
	) #Without time
	root = logging.getLogger()
	root.handlers = [handler]

	run_example_app()
